from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import next
from builtins import range
from builtins import object
import weakref

import pymel.util as util

# import all available Maya API methods in this module (api)
from maya.OpenMaya import *
from maya.OpenMayaAnim import *
try:
    from maya.OpenMayaCloth import *
except:
    pass
try:
    from maya.OpenMayaFX import *
except:
    pass
try:
    from maya.OpenMayaMPx import *
except:
    pass
if not MGlobal.mayaState() == MGlobal.kBatch:
    try:
        from maya.OpenMayaUI import *
    except:
        pass
try:
    from maya.OpenMayaRender import *
except:
    pass

# So, it seems that MScriptUtil().as*Ptr has a serious
# problem... basically, it looks like the returned
# ptrs don't have a reference to the MScriptUtil()
# that provides the actual storage for them -
# thus, it's possible for the MScriptUtil to get
# garbage collected, but still have the pointer
# to it in use - so it points to garbage (and
# generally causes a crash).
# To get around this, we create a wrapper for
# ptrs which also contains a reference to the
# MScriptUtil which contains the storage. Pass
# it around in place of the pointer - then, when
# the 'actual' pointer is needed (ie, immediately
# before feeding it into an api function), 'call'
# the SafeApiValue/Ptr object to return the 'true'
# pointer.
# Also, even SafeApiPtr is not completely safe -
# for instance, you cannot create 'throwaway' instances,
# like:
#   theApiFunc(SafeApiPtr('double')())
# ...as there is a chance that the MScriptUtil will be
# garbage collected before the api function tries to
# write into it's pointer...

# Note - I would have liked to have implemented this
# by simply attaching the MScriptUtil to the ptr -
# but you can't add attributes to the pointer object.
# My next idea was to use weakrefs to create a dictionary
# which maps ptrs to MScriptUtils - and clean it
# periodically as the ptrs are garbage collected. Alas,
# the pointer objects are also not compatible with
# weakref.  So, we have to use a 'non-transparent' wrapper...
# ie, we have to 'call' the object before feeding to
# the api function...


class SafeApiPtr(object):

    """
    A wrapper for api pointers which also contains a reference
    to the MScriptUtil which contains the storage. This helps
    ensure that the 'storage' for the pointer doesn't get
    wiped out before the pointer does. Pass the SafeApiPtr
    around in place of the 'true' pointer - then, when
    the 'true' pointer is needed (ie, immediately
    before feeding it into an api function), 'call'
    the SafeApiPtr object to return the 'true'
    pointer.

    Examples
    --------
    >>> from pymel.api.allapi import *
    >>> sel = MSelectionList()
    >>> sel.add('perspShape')
    >>> dag = MDagPath()
    >>> sel.getDagPath(0, dag)
    >>> cam = MFnCamera(dag)

    >>> aperMin = SafeApiPtr('double')
    >>> aperMax = SafeApiPtr('double')
    >>> cam.getFilmApertureLimits(aperMin(), aperMax())
    >>> print('%.5f, %.5f' % (aperMin.get(), aperMax.get()))
    0.01378, 20.28991
    """

    def __init__(self, valueType, scriptUtil=None, size=1, asTypeNPtr=False):
        # type: (str, MScriptUtil, int, bool) -> None
        """
        Parameters
        ----------
        valueType : str
            The name of the maya pointer type you would like
            returned - ie, 'int', 'short', 'float'.
        scriptUtil : `MScriptUtil`
            If you wish to use an existing MScriptUtil as
            the 'storage' for the value returned, specify it
            here - otherwise, a new MScriptUtil object is
            created.
        size : `int`
            If we want a pointer to an array, size indicates
            the number of items the array holds.  If we are
            creating an MScriptUtil, it will be initialized
            to hold this many items - if we are fed an
            MScriptUtil, then it is your responsibility to
            make sure it can hold the necessary number of items,
            or else maya will crash!
        asTypeNPtr : `bool`
            If we want a call to this SafeApiPtr to return a pointer
            for an argument such as:
               int2 &myArg;
            then we need to set asTypeNPtr to True:
               SafeApiPtr('int', size=2, asTypeNPtr=True)
            Otherwise, it is assumed that calling the object returns array
            ptrs:
               int myArg[2];
        """
        if not scriptUtil:
            self.scriptUtil = MScriptUtil()
            if size < 1:
                raise ValueError('size must be >= 1')
            else:
                # Value stored here doesn't matter - just make sure
                # it's large enough!
                self.scriptUtil.createFromList([0.0] * size, size)
        else:
            self.scriptUtil = scriptUtil
        self.size = size
        capValue = util.capitalize(valueType)
        self._normPtr = getattr(self.scriptUtil, 'as' + capValue + 'Ptr')()
        # Unforunately, arguments such as:
        #    float2 &foo;
        # need to be handled differently - calling it, we need
        # to return asFloat2Ptr()... but when indexing, use the same old
        # asFloatPtr() result to feed into getFloatArrayValue.
        # Also, note that asFloatPtr() must be called BEFORE asFloat2Ptr() -
        # if it is called after, the float2 ptr seems to get reset!
        self._sizedIndexGetter = None
        self._sizedIndexSetter = None
        if asTypeNPtr:
            self._nPtr = getattr(self.scriptUtil, 'as' + capValue +
                                 str(size) + 'Ptr')()
            self._ptr = self._nPtr
            self._sizedIndexGetter = getattr(
                MScriptUtil, 'get' + capValue + str(size) + 'ArrayItem', None)
            self._sizedIndexSetter = getattr(
                MScriptUtil, 'set' + capValue + str(size) + 'ArrayItem', None)
        else:
            self._ptr = self._normPtr
        self._getter = getattr(MScriptUtil, 'get' + capValue, None)
        self._setter = getattr(MScriptUtil, 'set' + capValue, None)
        self._indexGetter = getattr(MScriptUtil,
                                    'get' + capValue + 'ArrayItem', None)
        self._indexSetter = getattr(MScriptUtil,
                                    'set' + capValue + 'Array', None)

    def __call__(self):
        return self._ptr

    def get(self):
        """
        Dereference the pointer - ie, get the actual value we're pointing to.
        """
        return self._getter(self._normPtr)

    def set(self, value):
        """
        Store the actual value we're pointing to.
        """
        return self._setter(self._normPtr, value)

    def __getitem__(self, index):
        if index < 0 or index > (self.size - 1):
            raise IndexError(index)
        if self._sizedIndexGetter is not None:
            # as of 2018, MSCriptUtil won't return the right result if we don't
            # use, ie, getFloat2ArrayItem(nPtr, 0, 1) - just doing, ie,
            # getFloatArrayItem(normPtr, 1) won't work
            return self._sizedIndexGetter(self._nPtr, index // self.size,
                                          index % self.size)
        else:
            return self._indexGetter(self._normPtr, index)

    def __setitem__(self, index, value):
        if index < 0 or index > (self.size - 1):
            raise IndexError(index)
        if self._sizedIndexSetter is not None:
            return self._sizedIndexSetter(self._nPtr, index // self.size,
                                          index % self.size, value)
        else:
            return self._indexSetter(self._normPtr, index, value)

    def __len__(self):
        return self.size


# fast convenience tests on API objects
def isValidMObjectHandle(obj):
    if isinstance(obj, MObjectHandle):
        return obj.isValid() and obj.isAlive()
    else:
        return False


def isValidMObject(obj):
    if isinstance(obj, MObject):
        return not obj.isNull()
    else:
        return False


def isValidMPlug(obj):
    if isinstance(obj, MPlug):
        return not obj.isNull()
    else:
        return False


def isValidMDagPath(obj):
    if isinstance(obj, MDagPath):
        # when the underlying MObject is no longer valid, dag.isValid() will still return true,
        # but obj.fullPathName() will be an empty string
        return obj.isValid() and obj.fullPathName()
    else:
        return False


def isValidMNode(obj):
    if isValidMObject(obj):
        return obj.hasFn(MFn.kDependencyNode)
    else:
        return False


def isValidMDagNode(obj):
    if isValidMObject(obj):
        return obj.hasFn(MFn.kDagNode)
    else:
        return False


def isValidMNodeOrPlug(obj):
    return isValidMPlug(obj) or isValidMNode(obj)

# returns a MObject for an existing node


def toMObject(nodeName):
    """ Get the API MObject given the name of an existing node """
    sel = MSelectionList()
    obj = MObject()
    result = None
    try:
        sel.add(nodeName)
        sel.getDependNode(0, obj)
        if isValidMObject(obj):
            result = obj
    except:
        pass
    return result


def toApiObject(nodeName, plugs=True, comps=True, dagPlugs=True):
    # type: (Any, bool, bool, bool) -> None
    """ Get the API MPlug, MObject or (MObject, MComponent) tuple given the name
    of an existing node, attribute, components selection

    Parameters
    ----------
    plugs : bool
        if True, check if nodeName is an attribute/plug, and if it is, return
        a pair of (MDagPath, MPlug) if it's a dag node and dagPlugs is True,
        or just an MPlug if it's not a dag node or dagPlugs is False
    comps : bool
        if True, check if nodeName is a component name, and if it is, return
        a pair of (MDagPath, MObject), where the MObject contains the component.
    dagPlugs : bool
        if True, plug result will be a tuple of type (MDagPath, MPlug) for dag
        nodes

    If we were unable to retrieve the node/attribute/etc, returns None. None is
    also returned if name is of the form `validNodeName.invalidAttrOrComponent`,
    or if comps/plugs is False.  That is, if the given name is one that should
    be an attribute or component, it will either return None or the plug or
    component - it should never return just the node.
    """
    # special case check for empty string for speed...
    if not nodeName:
        return None

    # preferPlugs : Optional[bool]
    #     If both plugs and comps are True, and the given nodeName could yield
    #     EITHER a plug or component, then preferPlugs controls which is returned.
    #     If True, then the plug is returned; if False otherwise the component is.

    preferPlugs = True
    # Behavior of MSelectionList.getPlug changed in 2020.1,
    # where it would attempt to find a "related" plug when
    # a component was selected. (Previously, it just
    # errored.)
    import pymel.versions as versions
    if versions.current() >= versions.v2020_1:
        preferPlugs = False

    splitName = nodeName.split('.')
    sel = MSelectionList()
    try:
        sel.add(nodeName)
    except Exception:
        if len(splitName) > 1 and plugs:
            # Compound Attributes
            #  sometimes the index might be left off somewhere in a compound attribute
            # (ex 'Nexus.auxiliary.input' instead of 'Nexus.auxiliary[0].input' )
            #  but we can still get a representative plug. this will return the equivalent of 'Nexus.auxiliary[-1].input'
            try:
                obj = toApiObject(splitName[0], plugs=False, comps=False)
                if obj is None:
                    return None
                if isinstance(obj, MDagPath):
                    mfn = MFnDagNode(obj)
                else:
                    mfn = MFnDependencyNode(obj)
                plug = mfn.findPlug(splitName[-1], False)
                if isinstance(obj, MDagPath) and dagPlugs:
                    return (obj, plug)
                return plug
            except (RuntimeError, ValueError):
                pass
        return None

    if sel.length() != 1:
        return None

    if len(splitName) > 1:
        # early out - it's comp or plug name, but they didnt want either!
        if not (plugs or comps):
            return None

        def getAsPlug():
            plug = MPlug()
            try:
                sel.getPlug(0, plug)
            except RuntimeError:
                return None
            else:
                if dagPlugs and not plug.isNull():
                    # Check if it's a dag node - if so, want to return
                    # (MDagPath, MPlug) pair
                    try:
                        # Plugs with DagPaths
                        sel.add(splitName[0])
                        dag = MDagPath()
                        sel.getDagPath(1, dag)

                        # used to be a bug with some types that inherited from DagNode,
                        # but were not compatibile w/ MFnDagNode... these have been
                        # fixed, but we leave check in case another one crops up
                        if not dag.node().hasFn(MFn.kDagNode):
                            obj = MObject()
                            sel.getDependNode(1, obj)
                            return (obj, plug)

                        # if isValidMDagPath(dag) :
                        return (dag, plug)
                    except RuntimeError:
                        pass
            if plug.isNull():
                return None
            return plug

        def getAsComp():
            # Components
            dag = MDagPath()
            comp = MObject()
            try:
                sel.getDagPath(0, dag, comp)
            except RuntimeError:
                return None
            # if not isValidMDagPath(dag) :   return
            if comp.isNull():
                return None
            return (dag, comp)

        getters = []
        if plugs:
            getters.append(getAsPlug)
        if comps:
            getters.append(getAsComp)
        if not preferPlugs:
            getters.reverse()

        for getter in getters:
            result = getter()
            if result is not None:
                return result

        # We've failed to get either a plug or comp...

        # ...but we may have gotten a published container attribute, which
        # auto- magically converts to the contained node it references
        # when added to an MSelectionList
        if len(splitName) == 2:
            # Thankfully, it seems you can't index / get children off an
            # aliased attribute - ie, myNode.myAlias[0] and
            # myNode.myAlias.childAttr don't work, even if myAlias point
            # to a multi / compound attr
            obj = MObject()
            try:
                sel.add(splitName[0])
                sel.getDependNode(1, obj)
            except RuntimeError:
                pass
            else:
                # Since it seems there's no api way to get at the plug for
                # a published / aliased container attr, we just check for
                # aliases...
                mfn = MFnDependencyNode(obj)
                aliases = []
                if mfn.getAliasList(aliases):
                    for aliasName, trueName in util.pairIter(aliases):
                        if aliasName == splitName[1]:
                            return toApiObject(
                                '.'.join((splitName[0], trueName)),
                                plugs=plugs, comps=comps)
        # Ok, nothing worked (plug or comp), but we were given a plug/comp name.
        # Give up and return None
        return None
    else:
        # This is not a plug or component name!
        try:
            # DagPaths
            dag = MDagPath()
            sel.getDagPath(0, dag)
            # if not isValidMDagPath(dag) : return

            # used to be a bug with some types that inherited from DagNode,
            # but were not compatibile w/ MFnDagNode... these have been
            # fixed, but we leave check in case another one crops up
            if not dag.node().hasFn(MFn.kDagNode):
                raise RuntimeError
            return dag

        except RuntimeError:
            # Objects
            obj = MObject()
            sel.getDependNode(0, obj)
            # if not isValidMObject(obj) : return
            return obj


def toMDagPath(nodeName):
    """ Get an API MDagPAth to the node, given the name of an existing dag node """
    obj = toMObject(nodeName)
    if obj:
        dagFn = MFnDagNode(obj)
        dagPath = MDagPath()
        dagFn.getPath(dagPath)
        return dagPath


# returns a MPlug for an existing plug
def toMPlug(plugName):
    """ Get the API MObject given the name of an existing plug (node.attribute) """
    nodeAndAttr = plugName.split('.', 1)
    obj = toMObject(nodeAndAttr[0])
    plug = None
    if obj:
        depNodeFn = MFnDependencyNode(obj)
        attr = depNodeFn.attribute(nodeAndAttr[1])
        plug = MPlug(obj, attr)
        if plug.isNull():
            plug = None
    return plug


def toComponentMObject(dagPath):
    """
    get an MObject representing all components of the passed dagPath

    The component type that will be returned depends on the exact type of
    object passed in - for instance, a poly mesh will return a component
    representing all the kMeshVertComponents.

    The exact choice of component type is determined by MItGeometry.
    """

    component = MObject()
    sel = MSelectionList()
    mit = MItGeometry(dagPath)
    while not mit.isDone():
        # MItGeometry.component is deprecated
        comp = mit.currentItem()
        # merge is True
        sel.add(dagPath, comp, True)
        next(mit)
    sel.getDagPath(0, dagPath, component)
    return component


# MDagPath, MPlug or MObject to name
# Note there is a kNamedObject API type but not corresponding MFn, thus
# I see no way of querying the name of something that isn't a kDependency node or a MPlug
# TODO : add components support, short/ long name support where applies
def MObjectName(obj):
    """ Get the name of an existing MPlug, MDagPath or MObject representing a dependency node"""
    from builtins import str

    if isValidMPlug(obj):
        return obj.name()
    elif isValidMNode(obj):
        depNodeFn = MFnDependencyNode(obj)
        return depNodeFn.name()
    elif isValidMDagPath(obj):
        # return obj.fullPathName()
        return obj.partialPathName()
    else:
        return str(obj)


# names to MObjects function (expected to be faster to share one selectionList)
def nameToMObject(*args):
    """ Get the API MObjects given names of existing nodes """
    sel = MSelectionList()
    for name in args:
        sel.add(name)
    result = []
    obj = MObject()
    for i in range(sel.length()):
        try:
            sel.getDependNode(i, obj)
        except:
            result.append(None)
        if isValidMObject(obj):
            result.append(obj)
        else:
            result.append(None)
    if len(result) == 1:
        return result[0]
    else:
        return tuple(result)


# wrap of api iterators
def MItNodes(*args, **kwargs):
    """ Iterator on MObjects of nodes of the specified types in the Maya scene,
        if a list of tyes is passed as args, then all nodes of a type included in the list will be iterated on,
        if no types are specified, all nodes of the scene will be iterated on
        the types are specified as Maya API types """
    typeFilter = MIteratorType()
    if args:
        if len(args) == 1:
            typeFilter.setFilterType(args[0])
        else:
            # annoying argument conversion for Maya API non standard C types
            typeIntM = MIntArray()
            MScriptUtil.createIntArrayFromList(args, typeIntM)
            typeFilter.setFilterList(typeIntM)
        # we will iterate on dependancy nodes, not dagPaths or plugs
        typeFilter.setObjectType(MIteratorType.kMObject)
    # create iterator with (possibly empty) typeFilter
    iterObj = MItDependencyNodes(typeFilter)
    while not iterObj.isDone():
        yield (iterObj.thisNode())
        next(iterObj)


# Iterators on nodes connections using MItDependencyGraph (ie listConnections/ listHistory)
def MItGraph(nodeOrPlug, *args, **kwargs):
    """ Iterate over MObjects of Dependency Graph (DG) Nodes or Plugs starting at a specified root Node or Plug,
        If a list of types is provided, then only nodes of these types will be returned,
        if no type is provided all connected nodes will be iterated on.
        Types are specified as Maya API types.
        The following keywords will affect order and behavior of traversal:
        upstream: if True connections will be followed from destination to source,
                  if False from source to destination
                  default is False (downstream)
        breadth: if True nodes will be returned as a breadth first traversal of the connection graph,
                 if False as a preorder (depth first)
                 default is False (depth first)
        plug: if True traversal will be at plug level (no plug will be traversed more than once),
              if False at node level (no node will be traversed more than once),
              default is False (node level)
        prune : if True will stop the iteration on nodes than do not fit the types list,
                if False these nodes will be traversed but not returned
                default is False (do not prune) """
#    startObj = MObject()
#    startPlug = MPlug()
    startObj = None
    startPlug = None
    if isValidMPlug(nodeOrPlug):
        startPlug = nodeOrPlug
    elif isValidMNode(nodeOrPlug):
        startObj = nodeOrPlug
    else:
        raise ValueError("'%s' is not a valid Node or Plug" % MObjectName(nodeOrPlug))
    upstream = kwargs.get('upstream', False)
    breadth = kwargs.get('breadth', False)
    plug = kwargs.get('plug', False)
    prune = kwargs.get('prune', False)
    if args:
        typeFilter = MIteratorType()
        if len(args) == 1:
            typeFilter.setFilterType(args[0])
        else:
            # annoying argument conversion for Maya API non standard C types
            typeIntM = MIntArray()
            MScriptUtil.createIntArrayFromList(args, typeIntM)
            typeFilter.setFilterList(typeIntM)
        # we start on a node or a plug
        if startPlug is not None:
            typeFilter.setObjectType(MIteratorType.kMPlugObject)
        else:
            typeFilter.setObjectType(MIteratorType.kMObject)
    # create iterator with (possibly empty) filter list and flags
    if upstream:
        direction = MItDependencyGraph.kUpstream
    else:
        direction = MItDependencyGraph.kDownstream
    if breadth:
        traversal = MItDependencyGraph.kBreadthFirst
    else:
        traversal = MItDependencyGraph.kDepthFirst
    if plug:
        level = MItDependencyGraph.kPlugLevel
    else:
        level = MItDependencyGraph.kNodeLevel
    iterObj = MItDependencyGraph(startObj, startPlug, typeFilter, direction, traversal, level)
    if prune:
        iterObj.enablePruningOnFilter()
    else:
        iterObj.disablePruningOnFilter()
    # iterates and yields MObjects
    while not iterObj.isDone():
        yield (iterObj.thisNode())
        next(iterObj)


# Iterators on dag nodes hierarchies using MItDag (ie listRelatives)
def MItDag(root=None, *args, **kwargs):
    """ Iterate over the hierarchy under a root dag node, if root is None, will iterate on whole Maya scene
        If a list of types is provided, then only nodes of these types will be returned,
        if no type is provided all dag nodes under the root will be iterated on.
        Types are specified as Maya API types.
        The following keywords will affect order and behavior of traversal:
        breadth: if True nodes Mobjects will be returned as a breadth first traversal of the hierarchy tree,
                 if False as a preorder (depth first)
                 default is False (depth first)
        underworld: if True traversal will include a shape's underworld (dag object parented to the shape),
              if False underworld will not be traversed,
              default is False (do not traverse underworld )
        depth : if True will return depth of each node.
        prune : if True will stop the iteration on nodes than do not fit the types list,
                if False these nodes will be traversed but not returned
                default is False (do not prune) """
    # startObj = MObject()
    # startPath = MDagPath()
    startObj = startPath = None
    if isValidMDagPath(root):
        startPath = root
    elif isValidMDagNode(root):
        startObj = root
    else:
        raise ValueError("'%s' is not a valid Dag Node" % MObjectName(root))
    breadth = kwargs.get('breadth', False)
    underworld = kwargs.get('underworld', False)
    prune = kwargs.get('prune', False)
    path = kwargs.get('path', False)
    allPaths = kwargs.get('allPaths', False)
    if args:
        typeFilter = MIteratorType()
        if len(args) == 1:
            typeFilter.setFilterType(args[0])
        else:
            # annoying argument conversion for Maya API non standard C types
            typeIntM = MIntArray()
            MScriptUtil.createIntArrayFromList(args, typeIntM)
            typeFilter.setFilterList(typeIntM)
        # we start on a MDagPath or a Mobject
        if startPath is not None:
            typeFilter.setObjectType(MIteratorType.kMDagPathObject)
        else:
            typeFilter.setObjectType(MIteratorType.kMObject)
    # create iterator with (possibly empty) filter list and flags
    if breadth:
        traversal = MItDag.kBreadthFirst
    else:
        traversal = MItDag.kDepthFirst
    iterObj = MItDag(typeFilter, traversal)
    if root is not None:
        iterObj.reset(typeFilter, startObj, startPath, traversal)

    if underworld:
        iterObj.traverseUnderWorld(True)
    else:
        iterObj.traverseUnderWorld(False)
    # iterates and yields MObject or MDagPath
    # handle prune ?

    # Must define dPath in loop or the iterator will yield
    # them as several references to the same object (thus with the same value each time)
    # instances must not be returned multiple times
    # could use a dict but it requires "obj1 is obj2" and not only "obj1 == obj2" to return true to
    # dic = {}
    # dic[obj1]=True
    # dic.has_key(obj2)
    instance = []
    # code doesn't look nice but Im putting the tests out of the iter loops to loose as little speed as possible,
    # will certainly define functions for each case
    if allPaths:
        dPathArray = MDagPathArray()
        while not iterObj.isDone():
            if iterObj.isInstanced(True):
                obj = iterObj.currentItem()
                if not obj in instance:
                    iterObj.getAllPaths(dPathArray)
                    nbDagPath = dPathArray.length()
                    for i in range(nbDagPath):
                        dPath = MDagPath(dPathArray[i])
                        yield dPath
                    instance.append(obj)
            else:
                iterObj.getAllPaths(dPathArray)
                nbDagPath = dPathArray.length()
                for i in range(nbDagPath):
                    dPath = MDagPath(dPathArray[i])
                    yield dPath
            next(iterObj)
    elif path:
        while not iterObj.isDone():
            if iterObj.isInstanced(True):
                obj = iterObj.currentItem()
                if not obj in instance:
                    dPath = MDagPath()
                    iterObj.getPath(dPath)
                    yield dPath
                    instance.append(obj)
            else:
                dPath = MDagPath()
                iterObj.getPath(dPath)
                yield dPath
            next(iterObj)
    else:
        while not iterObj.isDone():
            obj = iterObj.currentItem()
            if iterObj.isInstanced(True):
                if not obj in instance:
                    yield obj
                    instance.append(obj)
            else:
                yield obj
            next(iterObj)


# Essentially duplicated in datatypes - only difference is
# whether return value is a PyMel or api object
# Repeated for speed
def getPlugValue(plug):
    """given an MPlug, get its value"""

    # if plug.isArray():
    #    raise TypeError, "array plugs of this type are not supported"

    obj = plug.attribute()
    apiType = obj.apiType()

    if apiType in [MFn.kAttribute2Double, MFn.kAttribute2Float, MFn.kAttribute2Short, MFn.kAttribute2Int,
                   MFn.kAttribute3Double, MFn.kAttribute3Float, MFn.kAttribute3Short, MFn.kAttribute3Int,
                   MFn.kAttribute4Double,
                   MFn.kCompoundAttribute]:
        res = []
        for i in range(plug.numChildren()):
            res.append(getPlugValue(plug.child(i)))
        return tuple(res)

    elif apiType in [MFn.kDoubleLinearAttribute, MFn.kFloatLinearAttribute]:
        return plug.asMDistance()

    elif apiType in [MFn.kDoubleAngleAttribute, MFn.kFloatAngleAttribute]:
        return plug.asMAngle()

    elif apiType == MFn.kTimeAttribute:
        return plug.asMTime()

    elif apiType == MFn.kNumericAttribute:
        # return getNumericPlugValue(plug)
        nAttr = MFnNumericAttribute(obj)
        dataType = nAttr.unitType()
        if dataType == MFnNumericData.kBoolean:
            return plug.asBool()

        elif dataType in [MFnNumericData.kShort, MFnNumericData.kInt, MFnNumericData.kLong, MFnNumericData.kByte]:
            return plug.asInt()

        elif dataType in [MFnNumericData.kFloat, MFnNumericData.kDouble, MFnNumericData.kAddr]:
            return plug.asDouble()
        raise "%s: unknown numeric attribute type: %s" % (plug.partialName(True, True, True, False, True, True), dataType)

    elif apiType == MFn.kEnumAttribute:
        return plug.asInt()

    elif apiType == MFn.kTypedAttribute:
        tAttr = MFnTypedAttribute(obj)
        dataType = tAttr.attrType()

        if dataType == MFnData.kInvalid:
            return None

        elif dataType == MFnData.kString:
            return plug.asString()

        elif dataType == MFnData.kNumeric:

            # all of the dynamic mental ray attributes fail here, but i have
            # no idea why they are numeric attrs and not message attrs.
            # cmds.getAttr returns None, so we will too.
            try:
                dataObj = plug.asMObject()
            except:
                return

            try:
                numFn = MFnNumericData(dataObj)
            except RuntimeError:
                if plug.isArray():
                    raise TypeError("%s: numeric arrays are not supported" %
                                    plug.partialName(True, True, True, False,
                                                     True, True))
                else:
                    raise TypeError("%s: attribute type is numeric, but its "
                                    "data cannot be interpreted numerically" %
                                    plug.partialName(True, True, True, False,
                                                     True, True))
            dataType = numFn.numericType()

            if dataType == MFnNumericData.kBoolean:
                return plug.asBool()

            elif dataType in [MFnNumericData.kShort, MFnNumericData.kInt, MFnNumericData.kLong, MFnNumericData.kByte]:
                return plug.asInt()

            elif dataType in [MFnNumericData.kFloat, MFnNumericData.kDouble, MFnNumericData.kAddr]:
                return plug.asDouble()

            elif dataType == MFnNumericData.k2Short:
                ptr1 = SafeApiPtr('short')
                ptr2 = SafeApiPtr('short')

                numFn.getData2Short(ptr1(), ptr2())
                return (ptr1.get(), ptr2.get())

            elif dataType in [MFnNumericData.k2Int, MFnNumericData.k2Long]:
                ptr1 = SafeApiPtr('int')
                ptr2 = SafeApiPtr('int')

                numFn.getData2Int(ptr1(), ptr2())
                return (ptr1.get(), ptr2.get())

            elif dataType == MFnNumericData.k2Float:
                ptr1 = SafeApiPtr('float')
                ptr2 = SafeApiPtr('float')

                numFn.getData2Float(ptr1(), ptr2())
                return (ptr1.get(), ptr2.get())

            elif dataType == MFnNumericData.k2Double:
                ptr1 = SafeApiPtr('double')
                ptr2 = SafeApiPtr('double')

                numFn.getData2Double(ptr1(), ptr2())
                return (ptr1.get(), ptr2.get())

            elif dataType == MFnNumericData.k3Float:
                ptr1 = SafeApiPtr('float')
                ptr2 = SafeApiPtr('float')
                ptr3 = SafeApiPtr('float')

                numFn.getData3Float(ptr1(), ptr2(), ptr3())
                return (ptr1.get(), ptr2.get(), ptr3.get())

            elif dataType == MFnNumericData.k3Double:
                ptr1 = SafeApiPtr('double')
                ptr2 = SafeApiPtr('double')
                ptr3 = SafeApiPtr('double')

                numFn.getData3Double(ptr1(), ptr2(), ptr3())
                return (ptr1.get(), ptr2.get(), ptr3.get())

            elif dataType == MFnNumericData.kChar:
                return plug.asChar()

            raise TypeError(
                "%s: Unsupported numeric attribute: %s" %
                (plug.partialName(True, True, True, False, True, True),
                 dataType))

        elif dataType == MFnData.kMatrix:
            return MFnMatrixData(plug.asMObject()).matrix()

        elif dataType == MFnData.kDoubleArray:
            try:
                dataObj = plug.asMObject()
            except RuntimeError:
                return []
            return MFnDoubleArrayData(dataObj).array()

        elif dataType == MFnData.kIntArray:
            try:
                dataObj = plug.asMObject()
            except RuntimeError:
                return []
            return MFnIntArrayData(dataObj).array()

        elif dataType == MFnData.kPointArray:
            try:
                dataObj = plug.asMObject()
            except RuntimeError:
                return []
            return MFnPointArrayData(dataObj).array()

        elif dataType == MFnData.kVectorArray:
            try:
                dataObj = plug.asMObject()
            except RuntimeError:
                return []
            return MFnVectorArrayData(dataObj).array()

        elif dataType == MFnData.kStringArray:
            try:
                dataObj = plug.asMObject()
            except RuntimeError:
                return []
            return MFnStringArrayData(dataObj).array()
        raise TypeError("%s: Unsupported typed attribute: %s" %
                        (plug.partialName(True, True, True, False, True, True),
                         dataType))

    raise TypeError("%s: Unsupported Type: %s" %
                    (plug.partialName(True, True, True, False, True, True),
                     obj.apiTypeStr()))




"""
pymel ipython configuration

Current Features
----------------

tab completion of depend nodes, dag nodes, and attributes
automatic import of pymel

Future Features
---------------

- tab completion of PyNode attributes
- color coding of tab complete options
    - to differentiate between methods and attributes
    - dag nodes vs depend nodes
    - shortNames vs longNames
- magic commands
- bookmarking of maya's recent project and files

To Use
------

place in your PYTHONPATH
add the following line to the 'main' function of $HOME/.ipython/ipy_user_conf.py::

    import ipymel

Author: Chad Dombrova
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from past.builtins import basestring
from builtins import object
from optparse import OptionParser
try:
    import maya
except ImportError as e:
    print("ipymel can only be setup if the maya package can be imported")
    raise e


import IPython

ipy_ver = IPython.__version__.split('.')
ipy_ver = [int(x) if x.isdigit() else x for x in ipy_ver]

if ipy_ver < [0, 11]:
    def get_ipython():
        import IPython.ipapi
        return IPython.ipapi.get()

    IPython.ipapi.IPApi.define_magic = IPython.ipapi.IPApi.expose_magic
    import IPython.ColorANSI as coloransi
    from IPython.genutils import page
    from IPython.ipapi import UsageError
    import IPython.Extensions.ipy_completers

    def get_colors(obj):
        return color_table[obj.rc.colors].colors
else:  # >= [0, 11]
    import IPython.utils.coloransi as coloransi
    from IPython.core.page import page
    from IPython.core.error import UsageError

    def get_colors(obj):
        return color_table[ip.colors].colors

if ipy_ver >= [0, 13]:
    def define_magic(interpreter, function):
        def get_ipython():
            return interpreter
        from IPython.core.magic import register_line_magic
        register_line_magic(function)
else:
    def define_magic(interpreter, function):
        interpreter.define_magic(function.__name__, function)

try:
    from IPython.core.error import TryNext
except ImportError:
    from IPython.ipapi import TryNext


Colors = coloransi.TermColors
ColorScheme = coloransi.ColorScheme
ColorSchemeTable = coloransi.ColorSchemeTable

ip = None

try:
    import readline
except ImportError:
    import pyreadline as readline
delim = readline.get_completer_delims()
delim = delim.replace('|', '')  # remove pipes
delim = delim.replace(':', '')  # remove colon
# delim = delim.replace("'", '') # remove quotes
# delim = delim.replace('"', '') # remove quotes
readline.set_completer_delims(delim)

import inspect
import re
import glob
import os
import shlex
import sys

# don't import pymel here, as this will trigger loading of maya/pymel
# immediately, and things in the userSetup.py won't get properly entered into
# the ipython shell's namespace... we need the startup of maya to happen
# from "within" ipython, ie, when we do:
#   ip.ex("from pymel.core import *")
# from pymel import core

# we also can't even use maya.cmds, because it doesn't work in anything other
# than the main thread... and most of the tab-completion stuff runs in a
# subthread... so api it is!
# Use api2 because it's faster...
import maya.api.OpenMaya as om

_scheme_default = 'Linux'

# Build a few color schemes
NoColor = ColorScheme(
    'NoColor', {
        'instance': Colors.NoColor,
        'collapsed': Colors.NoColor,
        'tree': Colors.NoColor,
        'transform': Colors.NoColor,
        'shape': Colors.NoColor,
        'nonunique': Colors.NoColor,
        'nonunique_transform': Colors.NoColor,

        'normal': Colors.NoColor  # color off (usu. Colors.Normal)
    })

LinuxColors = ColorScheme(
    'Linux', {
        'instance': Colors.LightCyan,
        'collapsed': Colors.Yellow,
        'tree': Colors.Green,
        'transform': Colors.White,
        'shape': Colors.LightGray,
        'nonunique': Colors.Red,
        'nonunique_transform': Colors.LightRed,

        'normal': Colors.Normal  # color off (usu. Colors.Normal)
    })

LightBGColors = ColorScheme(
    'LightBG', {
        'instance': Colors.Cyan,
        'collapsed': Colors.LightGreen,
        'tree': Colors.Blue,
        'transform': Colors.DarkGray,
        'shape': Colors.Black,
        'nonunique': Colors.Red,
        'nonunique_transform': Colors.LightRed,

        'normal': Colors.Normal  # color off (usu. Colors.Normal)
    })

# Build table of color schemes (needed by the dag_parser)
color_table = ColorSchemeTable([NoColor, LinuxColors, LightBGColors],
                               _scheme_default)
color_table['Neutral'] = LightBGColors


def splitDag(obj):
    buf = obj.split('|')
    tail = buf[-1]
    path = '|'.join(buf[:-1])
    return path, tail


def expand(obj):
    """
    allows for completion of objects that reside within a namespace. for example,
    ``tra*`` will match ``trak:camera`` and ``tram``

    for now, we will hardwire the search to a depth of three recursive namespaces.
    TODO:
    add some code to determine how deep we should go

    """
    return (obj + '*', obj + '*:*', obj + '*:*:*')


def api_ls(args, dagOnly, long=False):
    '''Because the tab completer runs in a subthread, and cmds.ls doesn't
    seem to work very well from a subthread, use maya.api.OpenMaya'''
    sel = om.MSelectionList()

    if isinstance(args, basestring):
        args = [args]

    for arg in args:
        # if it doesn't exist, MSelectionList.add will raise an error -
        # ignore that
        try:
            sel.add(arg)
        except Exception:
            pass
    if not long and not dagOnly:
        return list(sel.getSelectionStrings())

    # long is only used when getting nodes, not plugs, so ignore that case
    # for now...
    results = []
    mfnDep = om.MFnDependencyNode()
    for i in range(sel.length()):
        try:
            dagPath = sel.getDagPath(i)
        except TypeError:
            if dagOnly:
                continue
            mobj = sel.getDependNode(i)
            mfnDep.setObject(mobj)
            results.append(mfnDep.name())
        else:
            if long:
                results.append(dagPath.fullPathName())
            else:
                results.append(dagPath.partialPathName())
    return results


def api_children(path):
    sel = om.MSelectionList()
    try:
        sel.add(path)
    except RuntimeError:
        return []
    if not sel.length():
        return []
    try:
        dagPath = sel.getDagPath(0)
    except TypeError:
        return []
    return [om.MFnDagNode(dagPath.child(i)).fullPathName()
            for i in range(dagPath.childCount())]


def api_listAttr(path, shortNames=False):
    sel = om.MSelectionList()
    try:
        sel.add(path)
    except RuntimeError:
        return []
    if not sel.length():
        return []
    try:
        plug = sel.getPlug(0)
    except TypeError:
        try:
            node = om.MFnDependencyNode(sel.getDependNode(0))
        except RuntimeWarning:
            return []
        attrs = [om.MFnAttribute(node.attribute(i))
                 for i in range(node.attributeCount())]
        if shortNames:
            return [x.shortName for x in attrs]
        else:
            return [x.name for x in attrs]
    else:
        return [plug.child(i).partialName(useLongNames=not shortNames)
                for i in range(plug.numChildren())]


def complete_node_with_attr(node, attr):
    # print "noe_with_attr", node, attr
    long_attrs = api_listAttr(node)
    short_attrs = api_listAttr(node, shortNames=1)
    # if node is a plug  ( 'persp.t' ), the first result will be the passed plug
    if '.' in node:
        attrs = long_attrs[1:] + short_attrs[1:]
    else:
        attrs = long_attrs + short_attrs
    return [u'%s.%s' % (node, a) for a in attrs if a.startswith(attr)]


def pymel_dag_completer(self, event):
    return pymel_name_completer(self, event, dagOnly=True)


def pymel_name_completer(self, event, dagOnly=False):
    def get_children(obj, dagOnly):
        path, partialObj = splitDag(obj)
        # print "getting children", repr(path), repr(partialObj)

        # try:
        if True:
            fullpaths = api_ls(path, dagOnly, long=True)
            if not fullpaths or not fullpaths[0]:
                return []
            fullpath = fullpaths[0]
            children = api_children(fullpath)
            if not children:
                return []
        # except Exception:
        #     return []

        matchStr = fullpath + '|' + partialObj
        matches = [x.replace(fullpath, path, 1) for x in children if x.startswith(matchStr)]
        return matches

    # print "\nnode", repr(event.symbol), repr(event.line)
    # print "\nbegin"

    # note that the NAME_COMPLETER_RE also works for DAG_MAGIC_COMPLETER_RE
    # and DAG_COMPLETER_RE, since those are simply more restrictive versions,
    # which set "dagOnly"
    # print "text_until_cursor: {}".format(event.text_until_cursor)
    # print "symbol: {}".format(event.symbol)
    linematch = NAME_COMPLETER_RE.match(event.text_until_cursor)
    # print "linematch: {}".format(linematch.group(0))
    nametext = linematch.group('namematch')
    # print "nametext: {}".format(nametext)

    matches = None

    #--------------
    # Attributes
    #--------------
    if not dagOnly:
        attr_match = ATTR_RE.match(nametext)
    else:
        attr_match = None
    if attr_match:
        node, attr = attr_match.groups()
        if node == 'SCENE':
            res = api_ls(attr + '*', dagOnly)
            if res:
                matches = ['SCENE.' + x for x in res if '|' not in x]
        elif node.startswith('SCENE.'):
            node = node.replace('SCENE.', '')
            matches = ['SCENE.' + x for x in complete_node_with_attr(node, attr) if '|' not in x]
        else:
            matches = complete_node_with_attr(node, attr)

    #--------------
    # Nodes
    #--------------
    else:
        # we don't yet have a full node
        if '|' not in nametext or (nametext.startswith('|') and nametext.count('|') == 1):
            # print "partial node"
            kwargs = {}
            if nametext.startswith('|'):
                kwargs['long'] = True
            matches = api_ls(expand(nametext), dagOnly, **kwargs)

        # we have a full node, get it's children
        else:
            matches = get_children(nametext, dagOnly)

    if not matches:
        raise TryNext

    # if we have only one match, get the children as well
    if len(matches) == 1 and not attr_match:
        res = get_children(matches[0] + '|', dagOnly)
        matches += res

    if event.symbol != nametext:
        # in some situations, the event.symbol will only have incomplete
        # information - ie, if we are completing "persp|p", then the symbol will
        # be "p" - nametext will give us the full "persp|p", which we need so we
        # know we're checking for children of "persp". In these situations, we
        # need to STRIP the leading non-symbol portion, so we don't end up with
        # "persp|persp|perspShape" after completion.
        if nametext.endswith(event.symbol):
            if not event.symbol:
                preSymbol = nametext
            else:
                preSymbol = nametext[:-len(event.symbol)]
            matches = [x[len(preSymbol):] if x.startswith(preSymbol) else x
                       for x in matches]
        # HOWEVER - in other situations, the symbol will contain too much
        # information - ie, stuff that isn't strictly speaking a node name - such
        # as when we complete "SCENE.p".  In this case, the symbol is "SCENE.p",
        # whereas nametext is simply "p". In such cases, we need to PREPEND the
        # extra "SCENE." to the result, or else ipython will think our matches
        # are not actually matches...
        elif event.symbol.endswith(nametext):
            if not nametext:
                symbolPrefix = event.symbol
            else:
                symbolPrefix = event.symbol[:-len(nametext)]
            matches = [symbolPrefix + x for x in matches]
    return matches


PYTHON_TOKEN_RE = re.compile(r"(\S+(\.\w+)*)\.(\w*)$")


def pymel_python_completer(self, event):
    """Match attributes or global python names"""
    import pymel.core as pm

    # print "python_matches"
    text = event.symbol
    # print repr(text)
    # Another option, seems to work great. Catches things like ''.<tab>
    m = PYTHON_TOKEN_RE.match(text)

    if not m:
        raise TryNext

    expr, attr = m.group(1, 3)
    # print type(self.Completer), dir(self.Completer)
    # print self.Completer.namespace
    # print self.Completer.global_namespace
    try:
        # print "first"
        obj = eval(expr, self.Completer.namespace)
    except Exception:
        try:
            # print "second"
            obj = eval(expr, self.Completer.global_namespace)
        except Exception:
            raise TryNext
    # print "complete"
    if isinstance(obj, (pm.nt.DependNode, pm.Attribute)):
        # print "isinstance"
        node = str(obj)
        long_attrs = api_listAttr(node)
        short_attrs = api_listAttr(node, shortNames=1)

        matches = []
        matches = self.Completer.python_matches(text)
        # print "here"
        # if node is a plug  ( 'persp.t' ), the first result will be the passed plug
        if '.' in node:
            attrs = long_attrs[1:] + short_attrs[1:]
        else:
            attrs = long_attrs + short_attrs
        # print "returning"
        matches += [expr + '.' + at for at in attrs]
        #import colorize
        #matches = [ colorize.colorize(x,'magenta') for x in matches ]
        return matches

    raise TryNext


def buildRecentFileMenu():
    import pymel.core as pm

    if "RecentFilesList" not in pm.optionVar:
        return

    # get the list
    RecentFilesList = pm.optionVar["RecentFilesList"]
    nNumItems = len(RecentFilesList)
    RecentFilesMaxSize = pm.optionVar["RecentFilesMaxSize"]

#        # check if there are too many items in the list
#        if (RecentFilesMaxSize < nNumItems):
#
#            #if so, truncate the list
#            nNumItemsToBeRemoved = nNumItems - RecentFilesMaxSize
#
#            #Begin removing items from the head of the array (least recent file in the list)
#            for ($i = 0; $i < $nNumItemsToBeRemoved; $i++):
#
#                core.optionVar -removeFromArray "RecentFilesList" 0;
#
#            RecentFilesList = core.optionVar["RecentFilesList"]
#            nNumItems = len($RecentFilesList);

    # The RecentFilesTypeList optionVar may not exist since it was
    # added after the RecentFilesList optionVar. If it doesn't exist,
    # we create it and initialize it with a guess at the file type
    if nNumItems > 0:
        if "RecentFilesTypeList" not in pm.optionVar:
            pm.mel.initRecentFilesTypeList(RecentFilesList)

        RecentFilesTypeList = pm.optionVar["RecentFilesTypeList"]

    # toNativePath
    # first, check if we are the same.


def open_completer(self, event):
    relpath = event.symbol
    # print event # dbg
    if '-b' in event.line:
        # return only bookmark completions
        bkms = self.db.get('bookmarks', {})
        return list(bkms.keys())

    if event.symbol == '-':
        width_dh = str(len(str(len(ip.user_ns['_sh']) + 1)))
        # jump in directory history by number
        fmt = '-%0' + width_dh + 'd [%s]'
        ents = [fmt % (i, s) for i, s in enumerate(ip.user_ns['_sh'])]
        if len(ents) > 1:
            return ents
        return []

    raise TryNext


class TreePager(object):

    def __init__(self, colors, options):
        self.colors = colors
        self.options = options

    # print options.depth
    def do_level(self, obj, depth, isLast):
        if isLast[-1]:
            sep = '`-- '
        else:
            sep = '|-- '
        #sep = '|__ '
        depth += 1
        branch = ''
        for x in isLast[:-1]:
            if x:
                branch += '    '
            else:
                branch += '|   '
        branch = self.colors['tree'] + branch + sep + self.colors['normal']

        children = self.getChildren(obj)
        name = self.getName(obj)

        num = len(children) - 1

        if children:
            if self.options.maxdepth and depth >= self.options.maxdepth:
                state = '+'
            else:
                state = '-'
            pre = self.colors['collapsed'] + state + ' '
        else:
            pre = '  '

        yield pre + branch + name + self.colors['normal'] + '\n'
        # yield  Colors.Yellow + branch + sep + Colors.Normal+ name + '\n'

        if not self.options.maxdepth or depth < self.options.maxdepth:
            for i, x in enumerate(children):
                for line in self.do_level(x, depth, isLast + [i == num]):
                    yield line

    def make_tree(self, roots):
        num = len(roots) - 1
        tree = ''
        for i, x in enumerate(roots):
            for line in self.do_level(x, 0, [i == num]):
                tree += line
        return tree


class DagTree(TreePager):

    def getChildren(self, obj):
        if self.options.shapes:
            return obj.getChildren()
        else:
            return obj.getChildren(type='transform')

    def getName(self, obj):
        import pymel.core as pm
        name = obj.nodeName()

        if obj.isInstanced():
            if isinstance(obj, pm.nt.Transform):
                # keep transforms bolded
                color = self.colors['nonunique_transform']
            else:
                color = self.colors['nonunique']
            id = obj.instanceNumber()
            if id != 0:
                source = ' -> %s' % obj.getOtherInstances()[0]
            else:
                source = ''
            name = color + name + self.colors['instance'] + ' [' + str(id) + ']' + source
        elif not obj.isUniquelyNamed():
            if isinstance(obj, pm.nt.Transform):
                # keep transforms bolded
                color = self.colors['nonunique_transform']
            else:
                color = self.colors['nonunique']
            name = color + name
        elif isinstance(obj, pm.nt.Transform):
            # bold
            name = self.colors['transform'] + name
        else:
            name = self.colors['shape'] + name

        return name

# formerly: magic_dag
dag_parser = OptionParser()
dag_parser.add_option("-d", type="int", dest="maxdepth")
dag_parser.add_option("-t", action="store_false", dest="shapes", default=True)
dag_parser.add_option("-s", action="store_true", dest="shapes")


def dag(self, parameter_s=''):
    import pymel.core as pm

    options, args = dag_parser.parse_args(parameter_s.split())
    colors = get_colors(self)
    dagtree = DagTree(colors, options)
    if args:
        roots = [pm.PyNode(args[0])]
    else:
        roots = pm.ls(assemblies=1)
    page(dagtree.make_tree(roots))


class DGHistoryTree(TreePager):

    def getChildren(self, obj):
        source, dest = obj
        return source.node().listConnections(plugs=True, connections=True, source=True, destination=False, sourceFirst=True)

    def getName(self, obj):
        source, dest = obj
        name = "%s -> %s" % (source, dest)
        return name

    def make_tree(self, root):
        import pymel.core as pm
        roots = pm.listConnections(root, plugs=True, connections=True, source=True, destination=False, sourceFirst=True)
        return TreePager.make_tree(self, roots)

# formerly: magic_dghist
dg_parser = OptionParser()
dg_parser.add_option("-d", type="int", dest="maxdepth")
dg_parser.add_option("-t", action="store_false", dest="shapes", default=True)
dg_parser.add_option("-s", action="store_true", dest="shapes")


def dghist(self, parameter_s=''):
    """

    """
    import pymel.core as pm

    options, args = dg_parser.parse_args(parameter_s.split())
    if not args:
        print("must pass in nodes to display the history of")
        return

    colors = get_colors(self)
    dgtree = DGHistoryTree(colors, options)

    roots = [pm.PyNode(args[0])]

    page(dgtree.make_tree(roots))

# formerly: magic_open


def openf(self, parameter_s=''):
    """Change the current working directory.

    This command automatically maintains an internal list of directories
    you visit during your IPython session, in the variable _sh. The
    command %dhist shows this history nicely formatted. You can also
    do 'cd -<tab>' to see directory history conveniently.

    Usage:

      openFile 'dir': changes to directory 'dir'.

      openFile -: changes to the last visited directory.

      openFile -<n>: changes to the n-th directory in the directory history.

      openFile --foo: change to directory that matches 'foo' in history

      openFile -b <bookmark_name>: jump to a bookmark set by %bookmark
         (note: cd <bookmark_name> is enough if there is no
          directory <bookmark_name>, but a bookmark with the name exists.)
          'cd -b <tab>' allows you to tab-complete bookmark names.

    Options:

    -q: quiet.  Do not print the working directory after the cd command is
    executed.  By default IPython's cd command does print this directory,
    since the default prompts do not display path information.

    Note that !cd doesn't work for this purpose because the shell where
    !command runs is immediately discarded after executing 'command'."""

    parameter_s = parameter_s.strip()
    #bkms = self.shell.persist.get("bookmarks",{})

    oldcwd = os.getcwd()
    numcd = re.match(r'(-)(\d+)$', parameter_s)
    # jump in directory history by number
    if numcd:
        nn = int(numcd.group(2))
        try:
            ps = ip.ev('_sh[%d]' % nn)
        except IndexError:
            print('The requested directory does not exist in history.')
            return
        else:
            opts = {}
#        elif parameter_s.startswith('--'):
#            ps = None
#            fallback = None
#            pat = parameter_s[2:]
#            dh = self.shell.user_ns['_sh']
#            # first search only by basename (last component)
#            for ent in reversed(dh):
#                if pat in os.path.basename(ent) and os.path.isdir(ent):
#                    ps = ent
#                    break
#
#                if fallback is None and pat in ent and os.path.isdir(ent):
#                    fallback = ent
#
#            # if we have no last part match, pick the first full path match
#            if ps is None:
#                ps = fallback
#
#            if ps is None:
#                print "No matching entry in directory history"
#                return
#            else:
#                opts = {}

    else:
        # turn all non-space-escaping backslashes to slashes,
        # for c:\windows\directory\names\
        parameter_s = re.sub(r'\\(?! )', '/', parameter_s)
        opts, ps = self.parse_options(parameter_s, 'qb', mode='string')

    # jump to previous
    if ps == '-':
        try:
            ps = ip.ev('_sh[-2]' % nn)
        except IndexError:
            raise UsageError('%cd -: No previous directory to change to.')
#        # jump to bookmark if needed
#        else:
#            if not os.path.exists(ps) or opts.has_key('b'):
#                bkms = self.db.get('bookmarks', {})
#
#                if bkms.has_key(ps):
#                    target = bkms[ps]
#                    print '(bookmark:%s) -> %s' % (ps,target)
#                    ps = target
#                else:
#                    if opts.has_key('b'):
#                        raise UsageError("Bookmark '%s' not found.  "
#                              "Use '%%bookmark -l' to see your bookmarks." % ps)

    # at this point ps should point to the target dir
    if ps:
        ip.ex('openFile("%s", f=1)' % ps)
#            try:
#                os.chdir(os.path.expanduser(ps))
#                if self.shell.rc.term_title:
#                    #print 'set term title:',self.shell.rc.term_title  # dbg
#                    platutils.set_term_title('IPy ' + abbrev_cwd())
#            except OSError:
#                print sys.exc_info()[1]
#            else:
#                cwd = os.getcwd()
#                dhist = self.shell.user_ns['_sh']
#                if oldcwd != cwd:
#                    dhist.append(cwd)
#                    self.db['dhist'] = compress_dhist(dhist)[-100:]

#        else:
#            os.chdir(self.shell.home_dir)
#            if self.shell.rc.term_title:
#                platutils.set_term_title("IPy ~")
#            cwd = os.getcwd()
#            dhist = self.shell.user_ns['_sh']
#
#            if oldcwd != cwd:
#                dhist.append(cwd)
#                self.db['dhist'] = compress_dhist(dhist)[-100:]
#        if not 'q' in opts and self.shell.user_ns['_sh']:
#            print self.shell.user_ns['_sh'][-1]

# maya sets a sigint / ctrl-c / KeyboardInterrupt handler that quits maya -
# want to override this to get "normal" python interpreter behavior, where it
# interrupts the current python command, but doesn't exit the interpreter


def ipymel_sigint_handler(signal, frame):
    raise KeyboardInterrupt


def install_sigint_handler(force=False):
    import signal
    if force or signal.getsignal(signal.SIGINT) == ipymel_sigint_handler:
        signal.signal(signal.SIGINT, ipymel_sigint_handler)

# unfortunately, it seems maya overrides the SIGINT hook whenever a plugin is
# loaded...


def sigint_plugin_loaded_callback(*args):
    # from the docs, as of 2015 the args are:
    #   ( [ pathToPlugin, pluginName ], clientData )
    install_sigint_handler()

sigint_plugin_loaded_callback_id = None

DAG_MAGIC_COMPLETER_RE = re.compile(r"(?P<preamble>%dag\s+)(?P<namematch>(?P<previous_parts>([a-zA-Z0-9:_]*\|)*)(?P<current_part>[a-zA-Z0-9:_]*))$")
DAG_COMPLETER_RE = re.compile(r"(?P<preamble>((.+(\s+|\())|(SCENE\.))[^\w|:._]*)(?P<namematch>(?P<previous_parts>([a-zA-Z0-9:_]*\|)+)(?P<current_part>[a-zA-Z0-9:_]*))$")
NAME_COMPLETER_RE = re.compile(r"(?P<preamble>((.+(\s+|\())|(SCENE\.))[^\w|:._]*)(?P<namematch>(?P<previous_parts>([a-zA-Z0-9:_.]*(\.|\|))*)(?P<current_part>[a-zA-Z0-9:_]*))$")
ATTR_RE = re.compile(r"""(?P<prefix>[a-zA-Z_0-9|:.]+)\.(?P<partial_attr>\w*)$""")


def setup(shell):
    global ip
    if hasattr(shell, 'get_ipython'):
        ip = shell.get_ipython()
    else:
        ip = get_ipython()

    ip.set_hook('complete_command', pymel_python_completer, re_key="(?!{})".format(NAME_COMPLETER_RE.pattern))
    ip.set_hook('complete_command', pymel_dag_completer, re_key=DAG_MAGIC_COMPLETER_RE.pattern)
    ip.set_hook('complete_command', pymel_dag_completer, re_key=DAG_COMPLETER_RE.pattern)
    ip.set_hook('complete_command', pymel_name_completer, re_key=NAME_COMPLETER_RE.pattern)
    ip.set_hook('complete_command', open_completer, str_key="openf")

    ip.ex("from pymel.core import *")
    # stuff in __main__ is not necessarily in ipython's 'main' namespace... so
    # if the user has something in userSetup.py that he wants put in the
    # "interactive" namespace, it won't be - unless we do this:
    ip.ex('from __main__ import *')
    # if you don't want pymel imported into the main namespace, you can replace the above with something like:
    #ip.ex("import pymel as pm")

    define_magic(ip, openf)
    define_magic(ip, dag)
    define_magic(ip, dghist)

    # add projects
    ip.ex("""
import os.path
for _mayaproj in optionVar.get('RecentProjectsList', []):
    _mayaproj = os.path.join( _mayaproj, 'scenes' )
    if _mayaproj not in _dh:
        _dh.append(_mayaproj)""")

    # add files
    ip.ex("""
import os.path
_sh=[]
for _mayaproj in optionVar.get('RecentFilesList', []):
    if _mayaproj not in _sh:
        _sh.append(_mayaproj)""")

    # setup a handler for ctrl-c / SIGINT / KeyboardInterrupt, so maya / ipymel
    # doesn't quit
    install_sigint_handler(force=True)
    # unfortunately, when Mental Ray loads, it installs a new SIGINT handler
    # which restores the old "bad" behavior... need to install a plugin callback
    # to restore ours...
    global sigint_plugin_loaded_callback_id
    import pymel.core as pm
    if sigint_plugin_loaded_callback_id is None:
        sigint_plugin_loaded_callback_id = pm.api.MSceneMessage.addStringArrayCallback(
            pm.api.MSceneMessage.kAfterPluginLoad,
            sigint_plugin_loaded_callback)


def main():
    import IPython

    ipy_ver = IPython.__version__.split('.')
    ipy_ver = [int(x) if x.isdigit() else x for x in ipy_ver]

    if ipy_ver >= [1, 0]:
        import IPython.terminal.ipapp
        app = IPython.terminal.ipapp.TerminalIPythonApp.instance()
        app.initialize()
        setup(app.shell)
        app.start()
    elif ipy_ver >= [0, 11]:
        import IPython.frontend.terminal.ipapp
        app = IPython.frontend.terminal.ipapp.TerminalIPythonApp.instance()
        app.initialize()
        setup(app.shell)
        app.start()
    else:
        import IPython.Shell

        shell = IPython.Shell.start()
        setup(shell)
        shell.mainloop()

if __name__ == '__main__':
    main()



'''
This module wraps maya.cmds to accept special pymel arguments.

There are a number of pymel objects which must be converted to a "mel-friendly"
representation. For example, in versions prior to 2009, some mel commands (ie, getAttr) which expect
string arguments will simply reject custom classes, even if they have a valid string representation.
Another Example is mel's matrix inputs, which expects a flat list of 16 flaots, while pymel's Matrix has a more typical
4x4 representation.

If you're having compatibility issues with your custom classes when passing them to maya.cmds,
simply add a __melobject__ function that returns a mel-friendly result and pass them to pymel's wrapped commands.

The wrapped commands in this module are the starting point for any other pymel customizations.

'''
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from past.builtins import basestring
import inspect
import sys
import re
import os
import types

import pymel.util as util
import maya.cmds
import warnings

if False:
    from typing import *

__all__ = ['getMelRepresentation']
_thisModule = sys.modules[__name__]

# In Maya 2012+, the error is:
#   ValueError: No object matches name: foo.bar
objectErrorType = ValueError
objectErrorReg = re.compile(',?No object matches name: ,?(.*)$')


def _testDecorator(function):
    def newFunc(*args, **kwargs):
        print("wrapped function for %s" % function.__name__)
        return function(*args, **kwargs)
    newFunc.__name__ = function.__name__
    newFunc.__doc__ = function.__doc__
    return newFunc


def getCmdName(inFunc):
    # type: (types.FunctionType) -> str
    '''Use in place of inFunc.__name__ when inFunc could be a maya.cmds cmd

    handles stubFuncs
    '''
    cmdName = inFunc.__name__
    if cmdName == 'stubFunc':
        sourceFile = inspect.getsourcefile(inFunc)
        if (isinstance(sourceFile, basestring) and
                os.path.join('maya', 'app', 'commands') in sourceFile):
            # Here's where it gets tricky... this is a fairly big hack, highly
            # dependent on the exact implementation of maya.app.commands.stubFunc...
            freevars = inFunc.__code__.co_freevars
            # in python 2.5, tuples don't have index / find methods
            if not hasattr(freevars, 'index'):
                freevars = list(freevars)
            freeVarIndex = freevars.index('command')
            if freeVarIndex:
                raise ValueError('could not find a command var in %s' % cmdName)
            cmdName = inFunc.__closure__[freeVarIndex].cell_contents
    return cmdName


def getMelRepresentation(args, recursionLimit=None, maintainDicts=True):
    """Will return a list which contains each element of the iterable 'args' converted to a mel-friendly representation.

    :Parameters:
        recursionLimit : int or None
            If an element of args is itself iterable, recursionLimit specifies the depth to which iterable elements
            will recursively search for objects to convert; if ``recursionLimit==0``, only the elements
            of args itself will be searched for PyNodes -  if it is 1, iterables within args will have getMelRepresentation called
            on them, etc.  If recursionLimit==None, then there is no limit to recursion depth.

        maintainDicts : bool
            In general, all iterables will be converted to tuples in the returned copy - however, if maintainDicts==True,
            then iterables for which ``util.isMapping()`` returns True will be returned as dicts.

    """
    if recursionLimit:
        recursionLimit -= 1

    if maintainDicts and util.isMapping(args):
        newargs = dict(args)
        argIterable = iter(args.items())
        isList = False
    else:
        newargs = list(args)
        argIterable = enumerate(args)
        isList = True

    for index, value in argIterable:
        try:
            newargs[index] = value.__melobject__()
        except AttributeError:
            if ((not recursionLimit) or recursionLimit >= 0) and util.isIterable(value):
                # ...otherwise, recurse if not at recursion limit and  it's iterable
                newargs[index] = getMelRepresentation(value, recursionLimit, maintainDicts)
    if isList:
        newargs = tuple(newargs)
    return newargs


def _createFunction(func, oldname, newname):

    old_code = func.__code__
    # want to change the name, not just of the func, but of the underlying
    # code object - this makes it much easier to get useful information when
    # using cProfile
    # unfortunately, this isn't easy - have to get hacky...
    # ...we could do it with a big string and exec, but then we'd lose both
    # syntax highlighting, and file + lineno info...

    codeNewname = '%s_wrapped' % oldname

    if hasattr(old_code, 'replace'):
        # Future-proofing - 3.9 introduces a .replace() method
        new_code = old_code.replace(co_name=codeNewname)
    elif sys.version_info[:2] >= (3, 8):
        # Future-proofing: 3.8 adds co_posonlyargcount
        new_code = types.CodeType(
            old_code.co_argcount,
            old_code.co_posonlyargcount,
            old_code.co_kwonlyargcount,
            old_code.co_nlocals,
            old_code.co_stacksize,
            old_code.co_flags,
            old_code.co_code,
            old_code.co_consts,
            old_code.co_names,
            old_code.co_varnames,
            old_code.co_filename,
            codeNewname,
            old_code.co_firstlineno,
            old_code.co_lnotab,
            old_code.co_freevars,
            old_code.co_cellvars,
        )
    # 2to3: remove switch when python-3 only
    elif sys.version_info[0] >= 3:
        # Python 3 supports co_kwonlyargcount
        new_code = types.CodeType(
            old_code.co_argcount,
            old_code.co_kwonlyargcount,
            old_code.co_nlocals,
            old_code.co_stacksize,
            old_code.co_flags,
            old_code.co_code,
            old_code.co_consts,
            old_code.co_names,
            old_code.co_varnames,
            old_code.co_filename,
            codeNewname,
            old_code.co_firstlineno,
            old_code.co_lnotab,
            old_code.co_freevars,
            old_code.co_cellvars,
        )
    else:
        new_code = types.CodeType(
            old_code.co_argcount,
            old_code.co_nlocals,
            old_code.co_stacksize,
            old_code.co_flags,
            old_code.co_code,
            old_code.co_consts,
            old_code.co_names,
            old_code.co_varnames,
            old_code.co_filename,
            str(codeNewname),  # unicode no good in py2
            old_code.co_firstlineno,
            old_code.co_lnotab,
            old_code.co_freevars,
            old_code.co_cellvars,
        )
        newname = str(newname)  # unicode no good in py2

    return types.FunctionType(new_code,
                              func.__globals__,
                              newname,
                              func.__defaults__,
                              func.__closure__)


def addWrappedCmd(cmdname, cmd=None):
    # type: (str, Optional[types.FunctionType]) -> None
    if cmd is None:
        cmd = getattr(maya.cmds, cmdname)

    def wrappedCmd(*args, **kwargs):
        # we must get the cmd each time, because maya delays loading of functions until they are needed.
        # if we don't reload we'll keep the dummyFunc around
        new_cmd = getattr(maya.cmds, cmdname)
        # print args, kwargs
        # convert args to mel-friendly representation
        new_args = getMelRepresentation(args)

        # flatten list. this is necessary for list of components.
        # see Issue 71.  however, be sure that it's not an empty list/tuple
        if len(new_args) == 1 and util.isIterable(new_args[0]) and len(new_args[0]):
            new_args = new_args[0]

        new_kwargs = getMelRepresentation(kwargs)
        # print new_args, new_kwargs
        try:
            res = new_cmd(*new_args, **new_kwargs)
        except objectErrorType as e:
            # % formatter deals with unicode, but keeps str if not unicode
            m = objectErrorReg.match('%s' % e)
            if m:
                import pymel.core.general
                obj = m.group(1)
                raise pymel.core.general._objectError(obj)

            else:
                # re-raise error
                raise

        # when editing, some of maya.cmds functions return empty strings and
        # some return idiotic statements like 'Values Edited'.
        # however, for UI's in particular, people use the edit command to get
        # a pymel class for existing objects.
        # return None when we get an empty string
        try:
            if res == '' and kwargs.get('edit', kwargs.get('e', False)):
                return None
        except AttributeError:
            pass
        return res

    oldname = getattr(cmd, '__name__', None)
    if isinstance(oldname, basestring):
        # Don't use cmd.__name__, as this could be 'stubFunc'
        newname = getCmdName(cmd)
    else:
        newname = cmdname

    wrappedCmd = _createFunction(wrappedCmd, cmdname, newname)
    wrappedCmd.__doc__ = cmd.__doc__

    # for debugging, to make sure commands got wrapped...
    #wrappedCmd = _testDecorator(wrappedCmd)

    # so that we can identify that this is a wrapped maya command
    setattr(_thisModule, cmdname, wrappedCmd)


def removeWrappedCmd(cmdname):
    try:
        del cmdname
    except NameError:
        warnings.warn("%s not found in %s" % (cmdname, __name__))


def addAllWrappedCmds():
    for cmdname, cmd in inspect.getmembers(maya.cmds, callable):
        addWrappedCmd(cmdname, cmd)



"""
Functions and classes related to scripting, including `MelGlobals` and `Mel`
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import map
from builtins import range
from past.builtins import basestring
from builtins import object
from future.utils import PY2

import collections

# 2to3: remove switch when python-3 only
try:
    from collections.abc import Mapping, MutableMapping
except ImportError:
    from collections import Mapping, MutableMapping
import sys
import os
import inspect
from getpass import getuser as _getuser
from pymel.core import system

import maya.mel as _mm
import maya.cmds as _mc

import pymel.util as util
import pymel.internal.factories as _factories
import pymel.internal.cmdcache as _cmdcache
import pymel.core.datatypes as datatypes

if False:
    from typing import *
    from maya import cmds
    import maya.OpenMaya as _api
else:
    import pymel.api as _api
    import pymel.internal.pmcmds as cmds  # type: ignore[no-redef]

# -------------------------
# Mel <---> Python Glue
# -------------------------

MELTYPES = ['string', 'string[]', 'int', 'int[]', 'float', 'float[]', 'vector',
            'vector[]']


def isValidMelType(typStr):
    # type: (Any) -> bool
    """Returns whether ``typeStr`` is a valid MEL type identifier

    Returns
    -------
    bool
    """
    return typStr in MELTYPES


def _flatten(iterables):
    for it in iterables:
        if util.isIterable(it):
            for element in it:
                yield element
        else:
            yield it


def pythonToMel(arg):
    # type: (str) -> str
    """
    convert a python object to a string representing an equivalent value in mel

    iterables are flattened.

    mapping types like dictionaries have their key value pairs flattened:
        { key1 : val1, key2 : val2 }  -- >  ( key1, val1, key2, val2 )

    """
    if arg is None:
        return ''
    if arg is True or arg is False:
        return str(arg).lower()
    if util.isNumeric(arg):
        return str(arg)
    if isinstance(arg, datatypes.Vector):
        return '<<%f,%f,%f>>' % (arg[0], arg[1], arg[2])
    if util.isIterable(arg):
        if isinstance(arg, Mapping):
            arg = list(_flatten(arg.items()))
        else:
            arg = list(_flatten(arg))
        forceString = False
        for each in arg:
            if not util.isNumeric(each):
                forceString = True
                break

        if forceString:
            newargs = ['"%s"' % x for x in arg]
        else:
            newargs = [str(x) for x in arg]

        return '{%s}' % ','.join(newargs)

    # in order for PyNodes to get wrapped in quotes we have to treat special
    # cases first, we cannot simply test if arg is an instance of basestring
    # because PyNodes are not
    return '"%s"' % cmds.encodeString(str(arg))


def pythonToMelCmd(command, *args, **kwargs):
    '''Given a mel command name, and a set of python args / kwargs, return
    a mel string used to call the given command.

    Note that the first member of commandAndArgs is the command name, and is
    required; the remainder are the args to it.  We use this odd signature to
    avoid any name conflicts with mel flag names - ie, the signature used to be:
        pythonToMelCmd(command, *args, **kwargs)
    but this caused problems with the mel "button" function, which has a
    "command" flag.
    '''
    _factories.loadCmdCache()
    strArgs = [pythonToMel(arg) for arg in args]

    if kwargs:
        # keyword args trigger us to format as a command rather than a procedure
        strFlags = []  # type: List[str]
        if command in _factories.cmdlist:
            flags = _factories.cmdlist[command]['flags']
            shortFlags = _factories.cmdlist[command]['shortFlags']
        else:
            #            # Make a dummy flags dict - basically, just assume that q / e
            #            # are bool flags with no args...
            #            flags = {'query':{'args': bool,
            #                              'longname': 'query',
            #                              'numArgs': 0,
            #                              'shortname': 'q'},
            #                     'edit': {'args': bool,
            #                              'longname': 'edit',
            #                              'numArgs': 0,
            #                              'shortname': 'e'}}
            #            shortFlags = {'q':'query', 'e':'edit'}
            # Changed my mind - decided it's safest for unknown commands to
            # make no assumptions.  If they want a flag that takes no args,
            # they can use arg=None...
            flags = {}
            shortFlags = {}
        for key, val in kwargs.items():
            flagInfo = None
            if key in flags:
                flagInfo = flags[key]
            elif key in shortFlags:
                flagInfo = flags[shortFlags[key]]
            if (flagInfo and flagInfo.get('args') == bool
                    and flagInfo.get('numArgs') == 0):
                # we have a boolean argument that takes no args!
                # doing something like '-q 1' will raise an error, just
                # do '-q'
                strFlags.append('-%s' % key)
            elif (isinstance(val, (tuple, list))
                    and len(val) == flagInfo.get('numArgs')):
                strFlags.append('-%s %s' %
                                (key, ' '.join(pythonToMel(x) for x in val)))
            else:
                strFlags.append('-%s %s' % (key, pythonToMel(val)))
        cmdStr = '%s %s %s' % (command, ' '.join(strFlags), ' '.join(strArgs))
    else:
        # procedure
        cmdStr = '%s(%s)' % (command, ','.join(strArgs))
    return cmdStr


def getMelType(pyObj, exactOnly=True, allowBool=False, allowMatrix=False):
    # type: (Any, bool, bool, bool) -> str
    """
    return the name of the closest MEL type equivalent for the given python
    object.

    MEL has no true boolean or matrix types, but it often reserves special
    treatment for them in other ways.

    To control the handling of these types, use `allowBool` and `allowMatrix`.
    For python iterables, the first element in the array is used to determine
    the type. for empty lists, 'string[]' is returned.

        >>> from pymel.all import *
        >>> getMelType( 1 )
        'int'
        >>> p = SCENE.persp
        >>> getMelType( p.translate.get() )
        'vector'
        >>> getMelType( datatypes.Matrix )
        'int[]'
        >>> getMelType( datatypes.Matrix, allowMatrix=True )
        'matrix'
        >>> getMelType( True )
        'int'
        >>> getMelType( True, allowBool=True)
        'bool'
        >>> # make a dummy class
        >>> class MyClass(object): pass
        >>> getMelType( MyClass ) # returns None
        >>> getMelType( MyClass, exactOnly=False )
        'MyClass'

    Parameters
    ----------
    pyObj
        can be either a class or an instance.
    exactOnly : bool
        If True and no suitable MEL analog can be found, the function will
        return None.
        If False, types which do not have an exact mel analog will return
        the python type name as a string
    allowBool : bool
        if True and a bool type is passed, 'bool' will be returned.
        otherwise 'int'.
    allowMatrix : bool
        if True and a `Matrix` type is passed, 'matrix' will be returned.
        otherwise 'int[]'.

    Returns
    -------
    str
    """

    if inspect.isclass(pyObj):

        if issubclass(pyObj, basestring):
            return 'string'
        elif allowBool and issubclass(pyObj, bool):
            return 'bool'
        elif issubclass(pyObj, int):
            return 'int'
        elif issubclass(pyObj, float):
            return 'float'
        elif issubclass(pyObj, datatypes.VectorN):
            return 'vector'
        elif issubclass(pyObj, datatypes.MatrixN):
            if allowMatrix:
                return 'matrix'
            else:
                return 'int[]'

        elif not exactOnly:
            return pyObj.__name__

    else:
        if isinstance(pyObj, datatypes.VectorN):
            return 'vector'
        elif isinstance(pyObj, datatypes.MatrixN):
            if allowMatrix:
                return 'matrix'
            else:
                return 'int[]'
        elif util.isIterable(pyObj):
            try:
                return getMelType(pyObj[0], exactOnly=True) + '[]'
            except IndexError:
                # TODO : raise warning
                return 'string[]'
            except:
                return
        if isinstance(pyObj, basestring):
            return 'string'
        elif allowBool and isinstance(pyObj, bool):
            return 'bool'
        elif isinstance(pyObj, int):
            return 'int'
        elif isinstance(pyObj, float):
            return 'float'

        elif not exactOnly:
            return type(pyObj).__name__


# TODO : convert array variables to a semi-read-only list ( no append or extend, += is ok ):
# using append or extend will not update the mel variable

# we only inherit from dict for backward-compatability...
if False:
    _Parent = object
else:
    # 2to3: need to use "true" dict object in python-2 to avoid metaclass issue
    _Parent = __builtins__['dict']


class MelGlobals(MutableMapping, _Parent):

    """ A dictionary-like class for getting and setting global variables between mel and python.

    An instance of the class is created by default in the pymel namespace as ``melGlobals``.

    To retrieve existing global variables, just use the name as a key:

    >>> melGlobals['gPanelLabels'] #doctest: +ELLIPSIS
    [...'Side View', ...]
    >>> # works with or without $
    >>> melGlobals['$gFilterUIDefaultAttributeFilterList']  #doctest: +ELLIPSIS
    ['DefaultHiddenAttributesFilter', 'animCurveFilter', ..., 'publishedFilter']

    Creating new variables requires the use of the `initVar` function to specify the type:

    >>> melGlobals.initVar('string', 'gMyStrVar')
    '$gMyStrVar'
    >>> melGlobals['gMyStrVar'] = 'fooey'

    The variable will now be accessible within MEL as a global string.
    """
    melTypeToPythonType = {
        'string': str,
        'int': int,
        'float': float,
        'vector': datatypes.Vector
    }

#    class MelGlobalArray1( tuple ):
#        def __new__(cls, type, variable, *args, **kwargs ):
#
#            self = tuple.__new__( cls, *args, **kwargs )
#
#            decl_name = variable
#            if type.endswith('[]'):
#                type = type[:-2]
#                decl_name += '[]'
#
#            self._setItemCmd = "global %s %s; %s" % ( type, decl_name, variable )
#            self._setItemCmd += '[%s]=%s;'
#            return self
#
#        def setItem(self, index, value ):
#            _mm.eval(self._setItemCmd % (index, value) )

    class MelGlobalArray(util.defaultlist):

        def __init__(self, type, variable, *args, **kwargs):

            if type.endswith('[]'):
                type = type[:-2]

            pyType = MelGlobals.melTypeToPythonType[type]
            util.defaultlist.__init__(self, pyType, *args, **kwargs)

            declaration = MelGlobals._get_decl_statement(type, variable)
            self._setItemCmd = "%s; %s" % (declaration, variable)
            self._setItemCmd += '[%s]=%s;'

        def __setitem__(self, index, value):
            _mm.eval(self._setItemCmd % (index, pythonToMel(value)))
            super(MelGlobals.MelGlobalArray, self).__setitem__(index, value)
        setItem = __setitem__

        def append(self, val):
            raise AttributeError

        def extend(self, val):
            raise AttributeError

    typeMap = {}  # type: Dict[str, str]
    VALID_TYPES = MELTYPES

    def __iter__(self):
        for varName in mel.env():
            yield varName

    def __len__(self):
        return len(mel.env())

    def __getitem__(self, variable):
        # type: (str) -> str
        """
        Parameters
        ----------
        variable : str

        Returns
        -------
        str
        """
        return self.__class__.get(variable)

    def __setitem__(self, variable, value):
        # type: (str, Any) -> Any
        """
        Parameters
        ----------
        variable : str
        value : Any

        Returns
        -------
        Any
        """
        return self.__class__.set(variable, value)

    @classmethod
    def _formatVariable(cls, variable):
        # type: (str) -> str
        """
        Parameters
        ----------
        variable : str

        Returns
        -------
        str
        """
        # TODO : add validity check
        if not variable.startswith('$'):
            variable = '$' + variable
        if variable.endswith('[]'):
            variable = variable[:-2]
        return variable

    @classmethod
    def getType(cls, variable):
        # type: (str) -> str
        """
        Get the type of a global MEL variable

        Parameters
        ----------
        variable : str

        Returns
        -------
        str
        """
        variable = cls._formatVariable(variable)
        info = mel.whatIs(variable).split()
        if len(info) == 2 and info[1] == 'variable':
            MelGlobals.typeMap[variable] = info[0]
            return info[0]
        raise TypeError("Cannot determine type for this variable. "
                        "Use melGlobals.initVar first.")

    @classmethod
    def _get_decl_statement(cls, type, variable):
        decl_name = cls._formatVariable(variable)
        if type.endswith('[]'):
            type = type[:-2]
            decl_name += '[]'
        return "global %s %s" % (type, decl_name)

    @classmethod
    def initVar(cls, type, variable):
        # type: (str, str) -> str
        """
        Initialize a new global MEL variable

        Parameters
        ----------
        type : str
            one of ``MELTYPES``
        variable : str

        Returns
        -------
        str
        """
        if type not in MELTYPES:
            raise TypeError("type must be a valid mel type: %s" %
                            ', '.join(["'%s'" % x for x in MELTYPES]))
        variable = cls._formatVariable(variable)
        _mm.eval(cls._get_decl_statement(type, variable))
        MelGlobals.typeMap[variable] = type
        return variable

    def get_dict(self, variable, default=None):
        return super(MelGlobals, self).get(variable, default)

    # this clashes with a standard dict's "get", which will not error if a dict
    # does not contain a key...
    # ...but because of backwards compatibility, not sure what to do...?
    # for now, making a separate "get_default" method that acts like dict.get...
    # ...but may want to switch this in the future...
    @classmethod
    def get(cls, variable, type=None):
        # type: (str, Optional[str]) -> Any
        """
        get a MEL global variable.

        If the type is not specified, the mel ``whatIs`` command will be used
        to determine it.

        Parameters
        ----------
        variable : str
        type : Optional[str]
            one of ``MELTYPES``

        Returns
        -------
        Any
        """

        variable = cls._formatVariable(variable)
        if type is None:
            try:
                type = MelGlobals.typeMap[variable]
            except KeyError:
                try:
                    type = cls.getType(variable)
                except TypeError:
                    raise KeyError(variable)

        variable = cls.initVar(type, variable)

        if type.endswith('[]'):
            array = True
            proc_name = 'pymel_get_global_' + type.replace('[]', 'Array')
        else:
            array = False
            proc_name = 'pymel_get_global_' + type
        declaration = cls._get_decl_statement(type, variable)
        cmd = "global proc %s %s() { %s; return %s; } %s();" % (type, proc_name, declaration, variable, proc_name)
        # print cmd
        res = _mm.eval(cmd)
        if array:
            return MelGlobals.MelGlobalArray(type, variable, res)
        else:
            return MelGlobals.melTypeToPythonType[type](res)

    @classmethod
    def set(cls, variable, value, type=None):
        # type: (str, Any, Optional[str]) -> None
        """
        set a mel global variable

        Parameters
        ----------
        variable : str
        value : Any
        type : Optional[str]
            one of ``MELTYPES``
        """
        variable = cls._formatVariable(variable)
        if type is None:
            try:
                type = MelGlobals.typeMap[variable]
            except KeyError:
                type = cls.getType(variable)

        variable = cls.initVar(type, variable)
        declaration = cls._get_decl_statement(type, variable)
        cmd = "%s; %s=%s;" % (declaration, variable, pythonToMel(value))

        # print cmd
        _mm.eval(cmd)

    @classmethod
    def keys(cls):
        # type: () -> List[str]
        """
        list all global variables

        Returns
        -------
        List[str]
        """
        return mel.env()


melGlobals = MelGlobals()


# for backward compatibility
def getMelGlobal(type, variable):
    return melGlobals.get(variable, type)


def setMelGlobal(type, variable, value):
    return melGlobals.set(variable, value, type)


class Catch(object):

    """Reproduces the behavior of the mel command of the same name. if writing
    pymel scripts from scratch, you should use the try/except structure. This
    command is provided for python scripts generated by py2mel.  stores the
    result of the function in catch.result.

    >>> if not catch( lambda: myFunc( "somearg" ) ):
    ...    result = catch.result
    ...    print("succeeded:", result)

    """
    result = None
    success = None

    def __call__(self, func, *args, **kwargs):
        try:
            Catch.result = func(*args, **kwargs)
            Catch.success = True
            return 0
        except:
            Catch.success = False
            return 1

    def reset(self):
        Catch.result = None
        Catch.success = None


catch = Catch()


class OptionVarList(tuple):

    def __new__(cls, val, key):
        self = tuple.__new__(cls, val)
        return self

    def __init__(self, val, key):
        # tuple's init is object.__init__, which takes no args...
        #tuple.__init__(self, val)
        self.key = key

    def __setitem__(self, key, val):
        raise TypeError('%s object does not support item assignment - try '
                        'casting to a list, and assigning the whole list to '
                        'the optionVar' % self.__class__.__name__)

    def appendVar(self, val):
        """values appended to the OptionVarList with this method will be added
        to the Maya optionVar at the key denoted by self.key.
        """

        if isinstance(val, basestring):
            return cmds.optionVar(stringValueAppend=[self.key, val])
        if isinstance(val, (int, int)):
            return cmds.optionVar(intValueAppend=[self.key, val])
        if isinstance(val, float):
            return cmds.optionVar(floatValueAppend=[self.key, val])
        raise TypeError('unsupported datatype: strings, ints, floats and '
                        'their subclasses are supported')

    append = appendVar

# PY2: when we switch to python3-only, change
#    print(numArray)
# to
#    print(numArray)
# The int conversion is needed because in python-2, the values would be longs,
# ie, [1L, 24L, 7L, 9L]
class OptionVarDict(MutableMapping):

    """
    A dictionary-like class for accessing and modifying optionVars.

        >>> from pymel.all import *
        >>> optionVar['test'] = 'dooder'
        >>> optionVar['test']
        'dooder'

        >>> if 'numbers' not in env.optionVars:
        ...     optionVar['numbers'] = [1,24,7]
        >>> optionVar['numbers'].appendVar( 9 )
        >>> numArray = optionVar.pop('numbers')
        >>> print([int(x) for x in numArray])
        [1, 24, 7, 9]
        >>> optionVar.has_key('numbers') # previous pop removed the key
        False
    """

    def __call__(self, *args, **kwargs):
        return cmds.optionVar(*args, **kwargs)

    # use more efficient method provided by cmds.optionVar
    # (or at least, I hope it's more efficient...)
    def __contains__(self, key):
        return bool(cmds.optionVar(exists=key))

    # not provided by MutableMapping
    def has_key(self, key):
        return self.__contains__(key)

    def __getitem__(self, key):
        if key not in self:
            raise KeyError(key)
        val = cmds.optionVar(q=key)
        if isinstance(val, list):
            val = OptionVarList(val, key)
        return val

    def __setitem__(self, key, val):
        if isinstance(val, basestring):
            return cmds.optionVar(stringValue=[key, val])
        if isinstance(val, (int, bool, int)):
            return cmds.optionVar(intValue=[key, int(val)])
        if isinstance(val, float):
            return cmds.optionVar(floatValue=[key, val])
        if isinstance(val, (list, tuple)):
            if len(val) == 0:
                return cmds.optionVar(clearArray=key)
            listType = type(val[0])
            if issubclass(listType, basestring):
                flag = 'stringValue'
            elif issubclass(listType, (int, int)):
                flag = 'intValue'
            elif issubclass(listType, float):
                flag = 'floatValue'
            else:
                raise TypeError('%r is unsupported; Only strings, ints, float, lists, and their subclasses are supported' % listType)

            cmds.optionVar(**{flag: [key, val[0]]})  # force to this datatype
            flag += "Append"
            for elem in val[1:]:
                if not isinstance(elem, listType):
                    raise TypeError('all elements in list must be of the same datatype')
                cmds.optionVar(**{flag: [key, elem]})

    def keys(self):
        return cmds.optionVar(list=True)

    def pop(self, key):
        val = cmds.optionVar(q=key)
        cmds.optionVar(remove=key)
        return val

    def __delitem__(self, key):
        self.pop(key)

    def iterkeys(self):
        for key in self.keys():
            yield key
    __iter__ = iterkeys

    def __len__(self):
        return len(self.keys())


optionVar = OptionVarDict()


class Env(object):

    """ A Singleton class to represent Maya current optionVars and settings """

    optionVars = OptionVarDict()
    # grid = Grid()
    # playbackOptions = PlaybackOptions()

    # TODO : create a wrapper for os.environ which allows direct appending
    # and popping of individual env entries (i.e. make ':' transparent)
    envVars = os.environ

    def setConstructionHistory(self, state):
        cmds.constructionHistory(tgl=state)

    def getConstructionHistory(self):
        return cmds.constructionHistory(q=True, tgl=True)

    def sceneName(self):
        return system.Path(cmds.file(q=1, sn=1))

    def setUpAxis(self, axis, rotateView=False):
        """This flag specifies the axis as the world up direction. The valid
        axis are either 'y' or 'z'."""
        cmds.upAxis(axis=axis, rotateView=rotateView)

    def getUpAxis(self):
        """This flag gets the axis set as the world up direction. The valid
        axis are either 'y' or 'z'."""
        return cmds.upAxis(q=True, axis=True)

    def user(self):
        return _getuser()

    def host(self):
        return _gethostname()

    def getTime(self):
        return cmds.currentTime(q=1)

    def setTime(self, val):
        cmds.currentTime(val)
    time = property(getTime, setTime)

    def getMinTime(self):
        return cmds.playbackOptions(q=1, minTime=1)

    def setMinTime(self, val):
        cmds.playbackOptions(minTime=val)
    minTime = property(getMinTime, setMinTime)

    def getMaxTime(self):
        return cmds.playbackOptions(q=1, maxTime=1)

    def setMaxTime(self, val):
        cmds.playbackOptions(maxTime=val)
    maxTime = property(getMaxTime, setMaxTime)

    def getAnimStartTime(self):
        return cmds.playbackOptions(q=1, animationStartTime=1)

    def setAnimStartTime(self, val):
        cmds.playbackOptions(animationStartTime=val)
    animStartTime = property(getAnimStartTime, setAnimStartTime)

    def getAnimEndTime(self):
        return cmds.playbackOptions(q=1, animationEndTime=1)

    def setAnimEndTime(self, val):
        cmds.playbackOptions(animationEndTime=val)
    animEndTime = property(getAnimEndTime, setAnimEndTime)

    def getPlaybackTimes(self):
        return (self.animStartTime, self.minTime, self.maxTime,
                self.animEndTime)

    def setPlaybackTimes(self, playbackTimes):
        if len(playbackTimes) != 4:
            raise ValueError("must have 4 playback times")
        self.animStartTime = playbackTimes[0]
        self.minTime = playbackTimes[1]
        self.maxTime = playbackTimes[2]
        self.animEndTime = playbackTimes[3]
    playbackTimes = property(getPlaybackTimes, setPlaybackTimes)

env = Env()


# -------------------------
# Maya.mel Wrapper
# -------------------------

class MelError(RuntimeError):

    """Generic MEL error"""
    pass


class MelConversionError(MelError, TypeError):

    """MEL cannot process a conversion or cast between data types"""
    pass


class MelUnknownProcedureError(MelError, NameError):

    """The called MEL procedure does not exist or has not been sourced"""
    pass


class MelArgumentError(MelError, TypeError):

    """The arguments passed to the MEL script are incorrect"""
    pass


class MelSyntaxError(MelError, SyntaxError):

    """The MEL script has a syntactical error"""
    pass


class MelCallable(object):

    """
    Class for wrapping up callables created by Mel class' procedure calls.

    The class is designed to support chained, "namespace-protected" MEL
    procedure calls, like: ``Foo.bar.spam()``. In this case, Foo, bar and spam
    would each be `MelCallable` objects.
    """

    def __init__(self, head, name):
        if head:
            self.full_name = '%s.%s' % (head, name)
        else:
            self.full_name = name

    def __getattr__(self, command):
        if command.startswith('__') and command.endswith('__'):
            try:
                return self.__dict__[command]
            except KeyError:
                raise AttributeError("object has no attribute '%s'" % command)

        return MelCallable(head=self.full_name, name=command)

    def __call__(self, *args, **kwargs):
        cmd = pythonToMelCmd(self.full_name, *args, **kwargs)
        return Mel._eval(cmd, self.full_name)

# PY2: when we convert, remove the "#doctest: +IGNORE_EXCEPTION_DETAIL" bits
# They're needed, because in python 2, we get exceptions with no module:
#   MelConversionError: ...
# While in python 3, it includes the module:
#   pymel.core.language.MelConversionError: ...
# However, there's a known bug with ellipsis + doctest, so that this won't work:
#   ...MelConversionError: ...
# IGNORE_EXCEPTION_DETAIL is the ONLY thing that will make it ignore the leading
# module - however, that flag ALSO makes it completely ignore the exception
# value as well - all that is checked is the type (minus the module). So, once
# we can, get rid of IGNORE_EXCEPTION_DETAIL

class Mel(object):

    """Acts as a namespace from which MEL procedures can be called as if they
    were python functions.

    Automatically formats python arguments into a command string which is
    executed via ``maya.mel.eval``.  An instance of this class is created for
    you as `pymel.core.mel`.

    default:
        >>> import maya.mel as mel
        >>> # create the proc
        >>> mel.eval( 'global proc myScript( string $stringArg, float $floatArray[] ){}')
        >>> # run the script
        >>> mel.eval( 'myScript("firstArg", {1.0, 2.0, 3.0})')

    pymel:
        >>> from pymel.all import *
        >>> # create the proc
        >>> mel.eval( 'global proc myScript( string $stringArg, float $floatArray[] ){}')
        >>> # run the script
        >>> mel.myScript("firstArg", [1.0, 2.0, 3.0])

    The above is a very simplistic example. The advantages of pymel.mel over
    ``maya.mel.eval`` are more readily apparent when we want to pass a python
    object to our mel procedure:

    default:
        >>> import maya.cmds as cmds
        >>> node = "lambert1"
        >>> color = cmds.getAttr( node + ".color" )[0]
        >>> mel.eval('myScript("%s",{%f,%f,%f})' % (cmds.nodeType(node), color[0], color[1], color[2])   )

    pymel:
        >>> from pymel.all import *
        >>> node = PyNode("lambert1")
        >>> mel.myScript( node.type(), node.color.get() )

    In this you can see how `pymel.core.mel` allows you to pass any python
    object directly to your mel script as if it were a python function, with
    no need for formatting arguments.  The resulting code is much more readable.

    Another advantage of this class over maya.mel.eval is its handling of mel
    errors.  If a mel procedure fails to execute, you will get the specific mel
    error message in the python traceback, and, if they are enabled,
    line numbers!

    For example, in the example below we redeclare the myScript procedure with
    a line that will result in an error:

        >>> commandEcho(lineNumbers=1)  # turn line numbers on
        >>> mel.eval( '''
        ... global proc myScript( string $stringArg, float $floatArray[] ){
        ...     float $donuts = `ls -type camera`;}
        ... ''')
        >>> mel.myScript( 'foo', [] ) #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
          ...
        pymel.core.language.MelConversionError: Error during execution of MEL script: line 2: ,Cannot convert data of type string[] to type float.,
        Calling Procedure: myScript, in Mel procedure entered interactively.
          myScript("foo",{})

    Notice that the error raised is a `MelConversionError`.  There are several
    MEL exceptions that may be raised, depending on the type of error
    encountered: `MelError`, `MelConversionError`, `MelArgumentError`,
    `MelSyntaxError`, and `MelUnknownProcedureError`.

    Here's an example showing a `MelArgumentError`, which also demonstrates
    the additional traceback information that is provided for you, including
    the file of the calling script.

        >>> mel.startsWith('bar') #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
          ...
        pymel.core.language.MelArgumentError: Error during execution of MEL script: Line 1.18: ,Wrong number of arguments on call to startsWith.,
        Calling Procedure: startsWith, in file "..."
          startsWith("bar")

    Lastly, an example of `MelUnknownProcedureError`

        >>> mel.poop() #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
          ...
        pymel.core.language.MelUnknownProcedureError: Error during execution of MEL script: line 1: ,Cannot find procedure "poop".,

    Finally, some limitations: this Mel wrapper class cannot be used in
    situations in which the mel procedure modifies arguments (such as lists)
    in place, and you wish to then see the modified result in the calling code.
    Ie:

        >>> origList = []
        >>> newList = ["yet", "more", "things"]
        >>> mel.appendStringArray(origList, newList, 2)
        >>> origList
        []

    You will have to fall back to using mel.eval in such situations:

        >>> mel.eval('''
        ... string $origList[] = {};
        ... string $newList[] = {"yet", "more", "things"};
        ... appendStringArray($origList, $newList, 2);
        ... /* force a return value */
        ... $origList=$origList;
        ... ''')
        [u'yet', u'more']

    .. note::

        To remain backward compatible with maya.cmds, `MelError`, the base
        exception class that all these exceptions inherit from, in turn inherits
        from `RuntimeError`.
    """
    # proc is not an allowed name for a global procedure, so it's safe to use as an attribute
    proc = None

    def __getattr__(self, command):
        if command.startswith('__') and command.endswith('__'):
            try:
                return self.__dict__[command]
            except KeyError:
                raise AttributeError("object has no attribute '%s'" % command)

        return MelCallable(head='', name=command)

    @classmethod
    def mprint(cls, *args):
        """mel print command in case the python print command doesn't cut it"""
        # print r"""print (%s\\n);""" % pythonToMel( ' '.join( map( str, args)))
        _mm.eval( r"""print (%s);""" % pythonToMel( ' '.join( map( str, args))) + '\n' )

    @classmethod
    def source(cls, script, language='mel'):
        # type: (str, str) -> None
        """use this to source mel or python scripts.

        Parameters
        ----------
        script : str
        language : str
            {'mel', 'python'}
            When set to 'python', the source command will look for the
            python equivalent of this mel file, if it exists, and attempt
            to import it. This is particularly useful when transitioning
            from mel to python via `pymel.tools.mel2py`, with this simple
            switch you can change back and forth from sourcing mel to
            importing python.
        """

        if language == 'mel':
            cls.eval( """source "%s";""" % script )

        elif language == 'python':
            script = util.path(script)
            modulePath = script.namebase
            folder = script.parent
            print(modulePath)
            if modulePath not in sys.modules:
                print("importing")
                module = __import__(modulePath, globals(), locals(), [''])
                sys.modules[modulePath] = module

        else:
            raise TypeError("language keyword expects 'mel' or 'python'. "
                            "got '%s'" % language)

    @classmethod
    def eval(cls, cmd):
        """
        evaluate a string as a mel command and return the result.

        Behaves like `maya.mel.eval`, with several improvements:
            - returns `pymel.datatype.Vector` and `pymel.datatype.Matrix`
              classes
            - when an error is encountered a `MelError` exception is raised,
              along with the line number (if enabled) and exact mel error.

        >>> mel.eval( 'attributeExists("persp", "translate")' )
        0
        >>> mel.eval( 'interToUI( "fooBarSpangle" )' )
        'Foo Bar Spangle'

        """
        return cls._eval(cmd, None)

    @classmethod
    def _eval(cls, cmd, commandName):
        # commandName is just used for nicer formatting of error messages,
        # and is used by MelCallable

        # should return a value, like _mm.eval
        # return _mm.eval( cmd )
        # get this before installing the callback
        undoState = _mc.undoInfo(q=1, state=1)
        lineNumbers = _mc.commandEcho(q=1, lineNumbers=1)
        _mc.commandEcho(lineNumbers=1)
        global errors
        errors = []  # a list to store each error line

        def errorCallback(nativeMsg, messageType, data):
            global errors
            if messageType == _api.MCommandMessage.kError:
                if nativeMsg:
                    errors += [nativeMsg]

        # setup the callback:
        # assigning ids to a list avoids the swig memory leak warning, which would scare a lot of people even though
        # it is harmless.  hoping we get a real solution to this so that we don't have to needlessly accumulate this data
        id = _api.MCommandMessage.addCommandOutputCallback(errorCallback, None)

        try:
            res = _api.MCommandResult()
            _api.MGlobal.executeCommand(cmd, res, False, undoState)
        except Exception:
            msg = '\n'.join(errors)

            if 'Cannot find procedure' in msg:
                e = MelUnknownProcedureError
            elif 'Wrong number of arguments' in msg:
                e = MelArgumentError
                if commandName:
                    # remove the calling proc, it will be added below
                    msg = msg.split('\n', 1)[1].lstrip()
            elif 'Cannot convert data' in msg or 'Cannot cast data' in msg:
                e = MelConversionError
            elif 'Syntax error' in msg:
                e = MelSyntaxError
            else:
                e = MelError
            message = "Error during execution of MEL script: %s" % (msg)
            fmtCmd = '\n'.join(['  ' + x for x in cmd.split('\n')])

            if commandName:
                if e is not MelUnknownProcedureError:
                    file = _mm.eval('whatIs "%s"' % commandName)
                    if file.startswith('Mel procedure found in: '):
                        file = 'file "%s"' % os.path.realpath(file.split(':')[1].lstrip())
                    message += '\nCalling Procedure: %s, in %s' % (commandName, file)
                    message += '\n' + fmtCmd
            else:
                message += '\nScript:\n%s' % fmtCmd
            # PY2: once we switch to python-3 only, suppress the original
            # exception context, it's not useful
            #raise e(message) from None
            raise e(message)
        else:
            resType = res.resultType()

            if resType == _api.MCommandResult.kInvalid:
                return
            elif resType == _api.MCommandResult.kInt:
                result = _api.SafeApiPtr('int')
                res.getResult(result())
                return result.get()
            elif resType == _api.MCommandResult.kIntArray:
                result = _api.MIntArray()
                res.getResult(result)
                return [result[i] for i in range(result.length())]
            elif resType == _api.MCommandResult.kDouble:
                result = _api.SafeApiPtr('double')
                res.getResult(result())
                return result.get()
            elif resType == _api.MCommandResult.kDoubleArray:
                result = _api.MDoubleArray()
                res.getResult(result)
                return [result[i] for i in range(result.length())]
            elif resType == _api.MCommandResult.kString:
                return res.stringResult()
            elif resType == _api.MCommandResult.kStringArray:
                result = []
                res.getResult(result)
                return result
            elif resType == _api.MCommandResult.kVector:
                result = _api.MVector()
                res.getResult(result)
                return datatypes.Vector(result)
            elif resType == _api.MCommandResult.kVectorArray:
                result = _api.MVectorArray()
                res.getResult(result)
                return [datatypes.Vector(result[i]) for i in range(result.length())]
            elif resType == _api.MCommandResult.kMatrix:
                result = _api.MMatrix()
                res.getResult(result)
                return datatypes.Matrix(result)
            elif resType == _api.MCommandResult.kMatrixArray:
                result = _api.MMatrixArray()
                res.getResult(result)
                return [datatypes.Matrix(result[i]) for i in range(result.length())]
        finally:
            _api.MMessage.removeCallback(id)
            _mc.commandEcho(lineNumbers=lineNumbers)
            # 8.5 fix
            if hasattr(id, 'disown'):
                id.disown()

    @staticmethod
    def error(msg, showLineNumber=False):
        if showLineNumber:
            flags = ' -showLineNumber true '
        else:
            flags = ''
        _mm.eval( """error %s %s""" % ( flags, pythonToMel( msg) ) )

    @staticmethod
    def warning(msg, showLineNumber=False):
        if showLineNumber:
            flags = ' -showLineNumber true '
        else:
            flags = ''
        _mm.eval( """warning %s %s""" % ( flags, pythonToMel( msg) ) )

    @staticmethod
    def trace(msg, showLineNumber=False):
        if showLineNumber:
            flags = ' -showLineNumber true '
        else:
            flags = ''
        _mm.eval( """trace %s %s""" % ( flags, pythonToMel( msg) ) )

    @staticmethod
    def tokenize(*args):
        raise NotImplementedError("Calling the mel command 'tokenize' from "
                                  "python will crash Maya. Use the string "
                                  "split method instead.")

    # just a convenient alias
    globals = melGlobals


mel = Mel()


def conditionExists(conditionName):
    # type: (str) -> None
    """
    Returns True if the named condition exists, False otherwise.

    Parameters
    ----------
    conditionName : str
        A type used by `isTrue` and `scriptJob` (*Not* a type used by the
        condition *node*).
    """
    return conditionName in cmds.scriptJob(listConditions=True)


# ------ Do not edit below this line --------

evalEcho = _factories.getCmdFunc('evalEcho')

evalNoSelectNotify = _factories.getCmdFunc('evalNoSelectNotify')

getLastError = _factories.getCmdFunc('getLastError')

getProcArguments = _factories.getCmdFunc('getProcArguments')

matrixUtil = _factories.getCmdFunc('matrixUtil')

melOptions = _factories.getCmdFunc('melOptions')

optionVar = _factories.addCmdDocs(optionVar, cmdName='optionVar')

python = _factories.getCmdFunc('python')

resourceManager = _factories.getCmdFunc('resourceManager')

@_factories.addCmdDocs
def scriptJob(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['idleEvent', 'ie', 'tc', 'timeChange']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.scriptJob(*args, **kwargs)
    return res

sortCaseInsensitive = _factories.getCmdFunc('sortCaseInsensitive')

stackTrace = _factories.getCmdFunc('stackTrace')

waitCursor = _factories.getCmdFunc('waitCursor')



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import re

from future.utils import PY2

def trystr(input):
    '''If the input is unicode, but just holds normal ascii, convert to str'''
    if PY2:
        if isinstance(input, unicode):
            try:
                return str(input)
            except UnicodeEncodeError:
                pass
    return input

if PY2:
    RePattern = re._pattern_type
else:
    RePattern = re.Pattern


"""
Functions which are not listed in the maya documentation, such as commands created by plugins,
as well as the name parsing classes `DependNodeName`, `DagNodeName`, and `AttributeName`.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future.utils import PY2

from builtins import str
import re
import inspect

import pymel.internal.factories as _factories
if False:
    from maya import cmds
else:
    import pymel.internal.pmcmds as cmds  # type: ignore[no-redef]



# -------------------------
# Object Wrapper Classes
# -------------------------

class NameParser(str):
    PARENT_SEP = '|'

    def __new__(cls, strObj):
        """
        Casts a string to a pymel class.

        Use this function if you are unsure which class is the right one to use
        for your object.
        """
        strObj = str(strObj)
        # the if statement was failing for some types (ex: pymel.node.Vertex),
        # so forcing into unicode string:
        if cls is not NameParser:
            newcls = cls
        else:
            newcls = _getParserClass(strObj)
        self = super(NameParser, cls).__new__(newcls, strObj)
        return self

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__,
                           super(NameParser, self).__repr__())

    def __getattr__(self, attr):
        """
            >>> NameParser('foo:bar').spangle
            AttributeName('foo:bar.spangle')

        """
        if attr.startswith('__') and attr.endswith('__'):
            return super(NameParser, self).__getattr__(attr)

        return AttributeName('%s.%s' % (self, attr))

        #raise AttributeNameError, 'AttributeName does not exist %s' % attr

    def stripNamespace(self, levels=0):
        """
        Returns a new instance of the object with its namespace removed.  The calling instance is unaffected.
        The optional levels keyword specifies how many levels of cascading namespaces to strip, starting with the topmost (leftmost).
        The default is 0 which will remove all namespaces.

            >>> NameParser('foo:bar.spangle').stripNamespace()
            AttributeName('bar.spangle')

        """

        nodes = []
        for dagElem in self.split('|'):
            attrSplit = dagElem.split('.')
            spaceSplit = attrSplit[0].split(':')
            if levels:
                attrSplit[0] = ':'.join(spaceSplit[min(len(spaceSplit) - 1, levels):])
            else:
                attrSplit[0] = spaceSplit[-1]
            nodes.append('.'.join(attrSplit))
        return self.__class__('|'.join(nodes))

    def stripGivenNamespace(self, namespace, partial=True):
        """
        Returns a new instance of the object with any occurrences of the given namespace removed.  The calling instance is unaffected.
        The given namespace may end with a ':', or not.
        If partial is True (the default), and the given namespace has parent namespaces (ie, 'one:two:three'),
        then any occurrences of any parent namespaces are also stripped - ie, 'one' and 'one:two' would
        also be stripped.  If it is false, only namespaces

            >>> NameParser('foo:bar:top|foo:middle|foo:bar:extra:guy.spangle').stripGivenNamespace('foo:bar')
            AttributeName('top|middle|extra:guy.spangle')

            >>> NameParser('foo:bar:top|foo:middle|foo:bar:extra:guy.spangle').stripGivenNamespace('foo:bar', partial=False)
            AttributeName('top|foo:middle|extra:guy.spangle')
        """
        prefixSplit = namespace.rstrip(':').split(':')

        nodes = []
        for dagElem in self.split('|'):
            attrSplit = dagElem.split('.')
            spaceSplit = attrSplit[0].split(':')
            if partial:
                for toStrip in prefixSplit:
                    if spaceSplit[0] == toStrip:
                        spaceSplit.pop(0)
                    else:
                        break
            else:
                if spaceSplit[:len(prefixSplit)] == prefixSplit:
                    spaceSplit = spaceSplit[len(prefixSplit):]
                attrSplit[0] = spaceSplit[-1]
            attrSplit[0] = ':'.join(spaceSplit)
            nodes.append('.'.join(attrSplit))
        return self.__class__('|'.join(nodes))

    def swapNamespace(self, prefix):
        """
        Returns a new instance of the object with its current namespace
        replaced with the provided one.

        The calling instance is unaffected.
        """
        if not prefix.endswith(':'):
            prefix += ':'
        return self.__class__.addPrefix(self.stripNamespace(), prefix)

    def namespaceList(self):
        """
        Useful for cascading references.

        Returns all of the namespaces of the calling object as a list
        """
        return self.lstrip('|').rstrip('|').split('|')[-1].split(':')[:-1]

    def namespace(self):
        """
        Returns the namespace of the object with trailing colon included
        """
        nsList = self.namespaceList()
        if nsList:
            return ':'.join(nsList) + ':'
        return ''

    def addPrefix(self, prefix):
        'addPrefixToName'
        name = self
        leadingSlash = False
        if name.startswith('|'):
            name = name[1:]
            leadingSlash = True
        name = self.__class__('|'.join([prefix + x for x in name.split('|')]))
        if leadingSlash:
            name = '|' + name
        return self.__class__(name)

    def attr(self, attr):
        """access to AttributeName of a node. returns an instance of the
        AttributeName class for the given AttributeName.

            >>> NameParser('foo:bar').attr('spangle')
            AttributeName('foo:bar.spangle')

        """
        return AttributeName('%s.%s' % (self, attr))


class AttributeName(NameParser):

    """

    """
    attrItemReg = re.compile(r'\[(\d+:*\d*)\]$')

    # def __repr__(self):
    #    return "AttributeName('%s')" % self

    def __init__(self, attrName):
        attrName = str(attrName)
        if '.' not in attrName:
            raise TypeError("%s: AttributeNames must include the node and the "
                            "AttributeName. e.g. "
                            "'nodeName.AttributeNameName' " % self)
        self.__dict__['_multiattrIndex'] = 0

    def __getitem__(self, item):
        return AttributeName('%s[%s]' % (self, item))

    # Added the __call__ so to generate a more appropriate exception when a class method is not found
    def __call__(self, *args, **kwargs):
        raise TypeError("The object <%s> does not support the '%s' method" %
                        (repr(self.node()), self.plugAttr()))

    def array(self):
        """
        Returns the array (multi) AttributeName of the current element
            >>> n = AttributeName('lambert1.groupNodes[0]')
            >>> n.array()
            AttributeName('lambert1.groupNodes')
        """
        try:
            return AttributeName(AttributeName.attrItemReg.split(self)[0])
        except:
            raise TypeError("%s is not a multi AttributeName" % self)

    def plugNode(self):
        """plugNode

        >>> NameParser('foo:bar.spangle.banner').plugNode()
        DependNodeName('foo:bar')

        """
        return NameParser(str(self).split('.')[0])

    node = plugNode

    def plugAttr(self):
        """plugAttr

        >>> NameParser('foo:bar.spangle.banner').plugAttr()
        'spangle.banner'

        """
        if PY2:
            # just to get this method returning a builtins.str object, like
            # nearly everything else here does...
            return str('.').join(str(self).split('.')[1:])
        return '.'.join(str(self).split('.')[1:])

    def lastPlugAttr(self):
        """
        >>> NameParser('foo:bar.spangle.banner').lastPlugAttr()
        'banner'

        """
        return self.split('.')[-1]

    def nodeName(self):
        'basename'
        return self.split('|')[-1]

    def item(self, asSlice=False, asString=False):
        try:
            item = AttributeName.attrItemReg.search(self).group(1)
            if asString:
                return "[%s]" % str(item)
            val = item.split(":")
            val = [int(x) for x in val]
            if len(val) > 1:
                return asSlice and slice(*val) or val
            return val[0]
        except:
            return None

    def getParent(self, generations=1):
        """
        Returns the parent attribute

        Modifications:
            - added optional generations flag, which gives the number of levels up that you wish to go for the parent;
              ie:
                  >>> AttributeName("Cube1.multiComp[3].child.otherchild").getParent(2)
                  AttributeName('Cube1.multiComp[3]')

              Negative values will traverse from the top, not counting the initial node name:

                  >>> AttributeName("Cube1.multiComp[3].child.otherchild").getParent(-2)
                  AttributeName('Cube1.multiComp[3].child')

              A value of 0 will return the same node.
              The default value is 1.

              Since the original command returned None if there is no parent, to sync with this behavior, None will
              be returned if generations is out of bounds (no IndexError will be thrown).
        """

        if generations == 0:
            return self

        split = self.split('.')
        if -len(split) < generations < len(split) - 1:
            if generations < 0:
                # Move it one over to account for the initial node name
                splitIndex = 1 - generations
            else:
                splitIndex = -generations
            try:
                return AttributeName('.'.join(split[:splitIndex]))
            except:
                pass

    def addAttr(self, **kwargs):
        kwargs['longName'] = self.plugAttr()
        kwargs.pop('ln', None)
        from pymel.core.general import addAttr
        return addAttr(self.node(), **kwargs)

    def setAttr(self, *args, **kwargs):
        from pymel.core.general import setAttr
        return setAttr(self, *args, **kwargs)

    set = setAttr
    add = addAttr

    def exists(self):
        node = self.plugNode()
        attr = self.plugAttr()
        if not node or not attr:
            return False

        if not cmds.objExists(node):
            return False

        # short name
        if attr in cmds.listAttr(node, shortNames=1) + cmds.listAttr(node):
            return True

        return False


class DependNodeName(NameParser):
    # ------------------------------
    #    Name Info and Manipulation
    # ------------------------------

    def node(self):
        """
        for compatibility with AttributeName class
        """
        return self

    def nodeName(self):
        """
        for compatibility with DagNodeName class
        """
        return self

    def exists(self, **kwargs):
        "objExists"
        return bool(cmds.objExists(self, **kwargs))

    _numPartReg = re.compile('([0-9]+)$')

    def stripNum(self):
        """
        Return the name of the node with trailing numbers stripped off.

        If no trailing numbers are found the name will be returned unchanged."""
        try:
            return DependNodeName._numPartReg.split(self)[0]
        except IndexError:
            return str(self)

    def extractNum(self):
        """
        Return the trailing numbers of the node name.

        If no trailing numbers are found an error will be raised.
        """

        try:
            return DependNodeName._numPartReg.split(self)[1]
        except IndexError:
            raise ValueError("No trailing numbers to extract on "
                             "object %s" % self)

    def nextUniqueName(self):
        """
        Increment the trailing number of the object until a unique name is found

        If there is no trailing number, appends '1' to the name.
        Will always return a different name than the current name, even if the
            current name already does not exist.
        """
        try:
            name = self.nextName()
        except ValueError:
            name = self.__class__(self + '1')
        while name.exists():
            name = name.nextName()
        return name

    def nextName(self):
        """Increment the trailing number of the object by 1"""

        groups = DependNodeName._numPartReg.split(self)
        if len(groups) > 1:
            num = groups[1]
            formatStr = '%s%0' + str(len(num)) + 'd'
            return self.__class__(formatStr % (groups[0], (int(num) + 1)))
        else:
            raise ValueError("could not find trailing numbers to increment on "
                             "object %s" % self)

    def prevName(self):
        """Decrement the trailing number of the object by 1"""
        groups = DependNodeName._numPartReg.split(self)
        if len(groups) > 1:
            num = groups[1]
            formatStr = '%s%0' + str(len(num)) + 'd'
            return self.__class__(formatStr % (groups[0], (int(num) - 1)))
        else:
            raise ValueError("could not find trailing numbers to decrement on "
                             "object %s" % self)


class DagNodeName(DependNodeName):

    #    def __eq__(self, other):
    #        """ensures that we compare longnames when checking for dag node equality"""
    #        try:
    #            return unicode(self.longName()) == unicode(DagNodeName(other).longName())
    #        except (TypeError,IndexError):
    #            return unicode(self) == unicode(other)
    #
    #    def __ne__(self, other):
    #        """ensures that we compare longnames when checking for dag node equality"""
    #        try:
    #            return unicode(self.longName()) != unicode(DagNodeName(other).longName())
    #        except (TypeError,IndexError):
    #            return unicode(self) != unicode(other)

    # -------------------------
    #    DagNodeName Path Info
    # -------------------------
    def root(self):
        """rootOf"""
        return DagNodeName('|' + self.longName()[1:].split('|')[0])

    def getRoot(self):
        """unlike the root command which determines the parent via string formatting, this
        command uses the listRelatives command"""

        par = None
        cur = self
        while 1:
            par = cur.getParent()
            if not par:
                break
            cur = par
        return cur

    def firstParent(self):
        """firstParentOf"""

        return DagNodeName('|'.join(self.split('|')[:-1]))

    def getParent(self, generations=1):
        """
        Returns the parent node

        Modifications:
            - added optional generations flag, which gives the number of levels up that you wish to go for the parent;
              ie:
                  >>> DagNodeName("NS1:TopLevel|Next|ns2:Third|Fourth").getParent(2)
                  DagNodeName('NS1:TopLevel|Next')

              Negative values will traverse from the top, not counting the initial node name:

                  >>> DagNodeName("NS1:TopLevel|Next|ns2:Third|Fourth").getParent(-3)
                  DagNodeName('NS1:TopLevel|Next|ns2:Third')

              A value of 0 will return the same node.
              The default value is 1.

              Since the original command returned None if there is no parent, to sync with this behavior, None will
              be returned if generations is out of bounds (no IndexError will be thrown).
        """

        if generations == 0:
            return self

        split = self.split('|')
        if -len(split) <= generations < len(split):
            try:
                return DagNodeName('|'.join(split[:-generations]))
            except:
                pass


#    def shortName( self ):
#        'shortNameOf'
#        try:
#            return self.__class__( cmds.ls( self )[0] )
#        except:
#            return self

    def nodeName(self):
        """basename"""
        return self.split('|')[-1]


def _getParserClass(strObj):
    # First, see if strObj is actually a PyNode - in that case, get the class
    # based off the node...
    mro = set([cls.__name__ for cls in inspect.getmro(type(strObj))])
    # doing string comparison so we don't have to import core.general/nodetypes
    if 'PyNode' in mro:
        if 'DagNode' in mro:
            newcls = DagNodeName
        elif 'Attribute' in mro:
            newcls = AttributeName
        else:
            newcls = DependNodeName
    else:
        strObj = str(strObj)

        if '.' in strObj:
            newcls = AttributeName
            # Return Component Arrays ==========================================
            #            attr = obj.array().plugAttr()
            #            if attr in ["f","vtx","e","map"]:
            #                comps = getattr(Mesh(obj.node()), attr)
            #                return comps.__getitem__(obj.item(asSlice=1))
            #            else:
            #                return obj
            #===================================================================

        elif '|' in strObj:
            newcls = DagNodeName
        else:
            newcls = DependNodeName
    return newcls

# ------ Do not edit below this line --------

TanimLayer = _factories.getCmdFunc('TanimLayer')

TrenderSetupStates = _factories.getCmdFunc('TrenderSetupStates')

adpAnalyticsDialog = _factories.getCmdFunc('adpAnalyticsDialog')

adskAsset = _factories.getCmdFunc('adskAsset')

adskAssetLibrary = _factories.getCmdFunc('adskAssetLibrary')

adskAssetList = _factories.getCmdFunc('adskAssetList')

@_factories.addCmdDocs
def adskAssetListUI(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['cms', 'commandSuffix', 'uiC', 'uiCommand']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.adskAssetListUI(*args, **kwargs)
    return res

agFormatIn = _factories.getCmdFunc('agFormatIn')

agFormatOut = _factories.getCmdFunc('agFormatOut')

artAttr = _factories.getCmdFunc('artAttr')

artAttrSkinPaint = _factories.getCmdFunc('artAttrSkinPaint')

artAttrSkinPaintCmd = _factories.getCmdFunc('artAttrSkinPaintCmd')

artFluidAttr = _factories.getCmdFunc('artFluidAttr')

artSelect = _factories.getCmdFunc('artSelect')

artSetPaint = _factories.getCmdFunc('artSetPaint')

attrCompatibility = _factories.getCmdFunc('attrCompatibility')

blend = _factories.getCmdFunc('blend')

caddyManip = _factories.getCmdFunc('caddyManip')

clearShear = _factories.getCmdFunc('clearShear')

copyNode = _factories.getCmdFunc('copyNode')

crashInfoCmd = _factories.getCmdFunc('crashInfoCmd')

dagCommandWrapper = _factories.getCmdFunc('dagCommandWrapper')

dagObjectHit = _factories.getCmdFunc('dagObjectHit')

debug = _factories.getCmdFunc('debug')

debugNamespace = _factories.getCmdFunc('debugNamespace')

debugVar = _factories.getCmdFunc('debugVar')

dgControl = _factories.getCmdFunc('dgControl')

dgPerformance = _factories.getCmdFunc('dgPerformance')

dgcontrol = _factories.getCmdFunc('dgcontrol')

dgdebug = _factories.getCmdFunc('dgdebug')

dgstats = _factories.getCmdFunc('dgstats')

directConnectPath = _factories.getCmdFunc('directConnectPath')

dispatchGenericCommand = _factories.getCmdFunc('dispatchGenericCommand')

dynTestData = _factories.getCmdFunc('dynTestData')

evalContinue = _factories.getCmdFunc('evalContinue')

evaluationManagerInternal = _factories.getCmdFunc('evaluationManagerInternal')

evaluatorInternal = _factories.getCmdFunc('evaluatorInternal')

extendFluid = _factories.getCmdFunc('extendFluid')

@_factories.addCmdDocs
def flagTest(*args, **kwargs):
    for flag in ['timeRange', 'tr']:
        try:
            rawVal = kwargs[flag]
        except KeyError:
            continue
        else:
            kwargs[flag] = _factories.convertTimeValues(rawVal)
    res = cmds.flagTest(*args, **kwargs)
    return res

flushIdleQueue = _factories.getCmdFunc('flushIdleQueue')

flushThumbnailCache = _factories.getCmdFunc('flushThumbnailCache')

greasePencil = _factories.getCmdFunc('greasePencil')

greasePencilHelper = _factories.getCmdFunc('greasePencilHelper')

greaseRenderPlane = _factories.getCmdFunc('greaseRenderPlane')

groupParts = _factories.getCmdFunc('groupParts')

hotkeyEditor = _factories.getCmdFunc('hotkeyEditor')

hotkeyMapSet = _factories.getCmdFunc('hotkeyMapSet')

@_factories.addCmdDocs
def imageWindowEditor(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['cc', 'changeCommand']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.imageWindowEditor(*args, **kwargs)
    return res

interactionStyle = _factories.getCmdFunc('interactionStyle')

iterOnNurbs = _factories.getCmdFunc('iterOnNurbs')

journal = _factories.getCmdFunc('journal')

licenseCheck = _factories.getCmdFunc('licenseCheck')

manipComponentPivot = _factories.getCmdFunc('manipComponentPivot')

manipComponentUpdate = _factories.getCmdFunc('manipComponentUpdate')

matrix = _factories.getCmdFunc('matrix')

mayaDpiSettingAction = _factories.getCmdFunc('mayaDpiSettingAction')

meshIntersectTest = _factories.getCmdFunc('meshIntersectTest')

mimicMnipulation = _factories.getCmdFunc('mimicMnipulation')

mouldMesh = _factories.getCmdFunc('mouldMesh')

mouldSrf = _factories.getCmdFunc('mouldSrf')

mouldSubdiv = _factories.getCmdFunc('mouldSubdiv')

movieCompressor = _factories.getCmdFunc('movieCompressor')

myTestCmd = _factories.getCmdFunc('myTestCmd')

nodeGrapher = _factories.getCmdFunc('nodeGrapher')

nop = _factories.getCmdFunc('nop')

nurbsCurveRebuildPref = _factories.getCmdFunc('nurbsCurveRebuildPref')

ogsdebug = _factories.getCmdFunc('ogsdebug')

paint3d = _factories.getCmdFunc('paint3d')

polyColorSetCmdWrapper = _factories.getCmdFunc('polyColorSetCmdWrapper')

polyIterOnPoly = _factories.getCmdFunc('polyIterOnPoly')

polyPrimitiveMisc = _factories.getCmdFunc('polyPrimitiveMisc')

polySelectEditCtxDataCmd = _factories.getCmdFunc('polySelectEditCtxDataCmd')

polySelectSp = _factories.getCmdFunc('polySelectSp')

polySetVertices = _factories.getCmdFunc('polySetVertices')

polySpinEdge = _factories.getCmdFunc('polySpinEdge')

polyTestPop = _factories.getCmdFunc('polyTestPop')

polyToCurve = _factories.getCmdFunc('polyToCurve')

polyUVStackSimilarShellsCmd = _factories.getCmdFunc('polyUVStackSimilarShellsCmd')

polyWarpImage = _factories.getCmdFunc('polyWarpImage')

psdConvSolidTxOptions = _factories.getCmdFunc('psdConvSolidTxOptions')

rampWidget = _factories.getCmdFunc('rampWidget')

rampWidgetAttrless = _factories.getCmdFunc('rampWidgetAttrless')

readPDC = _factories.getCmdFunc('readPDC')

@_factories.addCmdDocs
def repeatLast(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['ac', 'acl', 'addCommand', 'addCommandLabel', 'cl', 'cnl', 'commandList', 'commandNameList']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.repeatLast(*args, **kwargs)
    return res

retimeHelper = _factories.getCmdFunc('retimeHelper')

safemodecheckhash = _factories.getCmdFunc('safemodecheckhash')

@_factories.addCmdDocs
def selectKeyframe(*args, **kwargs):
    for flag in ['t', 'time']:
        try:
            rawVal = kwargs[flag]
        except KeyError:
            continue
        else:
            kwargs[flag] = _factories.convertTimeValues(rawVal)
    res = cmds.selectKeyframe(*args, **kwargs)
    return res

subdDisplayMode = _factories.getCmdFunc('subdDisplayMode')

subdToNurbs = _factories.getCmdFunc('subdToNurbs')

subgraph = _factories.getCmdFunc('subgraph')

syncSculptCache = _factories.getCmdFunc('syncSculptCache')

testPa = _factories.getCmdFunc('testPa')

testPassContribution = _factories.getCmdFunc('testPassContribution')

texSculptCacheSync = _factories.getCmdFunc('texSculptCacheSync')

@_factories.addCmdDocs
def timeRangeInfo(*args, **kwargs):
    for flag in ['t', 'time']:
        try:
            rawVal = kwargs[flag]
        except KeyError:
            continue
        else:
            kwargs[flag] = _factories.convertTimeValues(rawVal)
    res = cmds.timeRangeInfo(*args, **kwargs)
    return res

timeSliderCustomDraw = _factories.getCmdFunc('timeSliderCustomDraw')

warnUserDialog = _factories.getCmdFunc('warnUserDialog')
